from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import httpx

from .exceptions import AuthenticationError, ServerNotReachableError

if TYPE_CHECKING:
    from .client import InfrahubClient, InfrahubClientSync


class ObjectStoreBase:
    pass


class ObjectStore(ObjectStoreBase):
    def __init__(self, client: InfrahubClient):
        self.client = client

    async def get(self, identifier: str, tracker: str | None = None) -> str:
        url = f"{self.client.address}/api/storage/object/{identifier}"
        headers = copy.copy(self.client.headers or {})
        if self.client.insert_tracker and tracker:
            headers["X-Infrahub-Tracker"] = tracker

        try:
            resp = await self.client._get(url=url, headers=headers)
            resp.raise_for_status()

        except ServerNotReachableError:
            self.client.log.error(f"Unable to connect to {self.client.address} .. ")
            raise
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code in [401, 403]:
                response = exc.response.json()
                errors = response.get("errors")
                messages = [error.get("message") for error in errors]
                raise AuthenticationError(" | ".join(messages)) from exc

        return resp.text

    async def upload(self, content: str, tracker: str | None = None) -> dict[str, str]:
        url = f"{self.client.address}/api/storage/upload/content"
        headers = copy.copy(self.client.headers or {})
        if self.client.insert_tracker and tracker:
            headers["X-Infrahub-Tracker"] = tracker

        try:
            resp = await self.client._post(url=url, payload={"content": content}, headers=headers)
            resp.raise_for_status()
        except ServerNotReachableError:
            self.client.log.error(f"Unable to connect to {self.client.address} .. ")
            raise
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code in [401, 403]:
                response = exc.response.json()
                errors = response.get("errors")
                messages = [error.get("message") for error in errors]
                raise AuthenticationError(" | ".join(messages)) from exc

        return resp.json()


class ObjectStoreSync(ObjectStoreBase):
    def __init__(self, client: InfrahubClientSync):
        self.client = client

    def get(self, identifier: str, tracker: str | None = None) -> str:
        url = f"{self.client.address}/api/storage/object/{identifier}"
        headers = copy.copy(self.client.headers or {})
        if self.client.insert_tracker and tracker:
            headers["X-Infrahub-Tracker"] = tracker

        try:
            resp = self.client._get(url=url, headers=headers)
            resp.raise_for_status()

        except ServerNotReachableError:
            self.client.log.error(f"Unable to connect to {self.client.address} .. ")
            raise
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code in [401, 403]:
                response = exc.response.json()
                errors = response.get("errors")
                messages = [error.get("message") for error in errors]
                raise AuthenticationError(" | ".join(messages)) from exc

        return resp.text

    def upload(self, content: str, tracker: str | None = None) -> dict[str, str]:
        url = f"{self.client.address}/api/storage/upload/content"
        headers = copy.copy(self.client.headers or {})
        if self.client.insert_tracker and tracker:
            headers["X-Infrahub-Tracker"] = tracker

        try:
            resp = self.client._post(url=url, payload={"content": content}, headers=headers)
            resp.raise_for_status()
        except ServerNotReachableError:
            self.client.log.error(f"Unable to connect to {self.client.address} .. ")
            raise
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code in [401, 403]:
                response = exc.response.json()
                errors = response.get("errors")
                messages = [error.get("message") for error in errors]
                raise AuthenticationError(" | ".join(messages)) from exc

        return resp.json()



from infrahub.core.diff.enricher.summary_counts import DiffSummaryCountsEnricher
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffSummaryCountsEnricherDependency(DependencyBuilder[DiffSummaryCountsEnricher]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffSummaryCountsEnricher:  # noqa: ARG003
        return DiffSummaryCountsEnricher()



import hashlib
import json
from datetime import datetime, timezone
from types import TracebackType
from typing import Any

import httpx
import pytest
from typing_extensions import Self

from .constants import PERFORMANCE_TEST_KIND, PERFORMANCE_TEST_VERSION
from .helpers import InfrahubDockerCompose
from .host import get_system_stats
from .models import (
    ContextUnit,
    InfrahubActiveMeasurementItem,
    InfrahubMeasurementItem,
    InfrahubResultContext,
    MeasurementDefinition,
)


class InfrahubPerformanceTest:
    context: dict[str, InfrahubResultContext]
    measurements: list[InfrahubMeasurementItem]
    active_measurements: InfrahubActiveMeasurementItem | None = None

    def __init__(self, results_url: str) -> None:
        self.name: str | None = None
        self.infrahub_version: str | None = None
        self.context = {}
        self.measurements = []
        self.metrics = {}
        self.host = get_system_stats()
        self.env_vars = {}
        self.project_name = ""
        self.test_info = {}
        self.start_time = datetime.now(timezone.utc)
        self.end_time: datetime | None = None
        self.results_url = results_url
        self.scraper_endpoint = ""
        self.initialized = False

    def initialize(
        self,
        name: str,
        compose: InfrahubDockerCompose | None = None,
        client: Any | None = None,  # noqa: ANN401
    ) -> None:
        self.name = name
        if client:
            self.infrahub_version = client.get_version()
        if compose:
            self.extract_compose_information(compose)

        self.initialized = True

    def finalize(self, session: pytest.Session) -> None:
        if self.initialized:
            self.end_time = datetime.now(timezone.utc)
            self.extract_test_session_information(session)
            self.send_results()

    def extract_compose_information(self, compose: InfrahubDockerCompose) -> None:
        self.env_vars = compose.env_vars
        self.project_name = compose.project_name
        self.scraper_endpoint = (
            f"http://127.0.0.1:{compose.get_service_host_and_port(service_name='scraper')[1]}/api/v1/export"
        )

    def extract_test_session_information(self, session: pytest.Session) -> None:
        self.test_info = {
            "summary": {
                "exitstatus": session.exitstatus,
                "nbr_tests": session.testscollected,
                "nbr_errors": session.testsfailed,
            }
        }

    def add_context(self, name: str, value: float | str = 0, unit: ContextUnit = ContextUnit.COUNT) -> None:
        self.context[name] = InfrahubResultContext(name=name, value=value, unit=unit)

    def add_measurement(
        self,
        definition: MeasurementDefinition,
        value: float | str,
        **kwargs: str | float,
    ) -> None:
        self.measurements.append(
            InfrahubMeasurementItem(
                name=definition.name,
                value=value,
                unit=definition.unit,
                context=kwargs or {},
            )
        )

    def start_measurement(self, definition: MeasurementDefinition, **kwargs: str | float) -> Self:
        self.active_measurements = InfrahubActiveMeasurementItem(definition=definition, context=kwargs or {})
        return self

    def fetch_metrics(self) -> None:
        with httpx.Client(timeout=30.0) as client:
            # Get Infrahub metrics
            response = client.post(
                url=self.scraper_endpoint,
                content='match[]={__name__=~"infrahub.*"}',
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
            self.metrics = [json.loads(line) for line in response.text.splitlines()]

            # Get system metrics, filter by docker project name
            response = client.post(
                url=self.scraper_endpoint,
                content=f'match[]={{__name__=~"container.*", container_label_com_docker_compose_project="{self.project_name}"}}',
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
            self.metrics += [json.loads(line) for line in response.text.splitlines()]

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if not exc_type and self.active_measurements:
            self.add_measurement(
                definition=self.active_measurements.definition,
                value=(datetime.now(timezone.utc) - self.active_measurements.start_time).total_seconds() * 1000,
                context=self.active_measurements.context,
            )

        # Don't record the measurement if the test failed
        self.active_measurements = None

    def _get_payload(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "infrahub_version": self.infrahub_version,
            "start_time": self.start_time.isoformat(),
            "end_time": self.end_time.isoformat() if self.end_time else None,
            "duration": (self.end_time - self.start_time).total_seconds() * 1000 if self.end_time else None,
            "context": [item.model_dump() for item in self.context.values()],
            "measurements": [item.model_dump() for item in self.measurements],
            "metrics": self.metrics,
            "host": self.host,
            "env_vars": self.env_vars,
            "test_info": self.test_info,
        }

    def send_results(self) -> None:
        data = self._get_payload()

        payload = {
            "kind": PERFORMANCE_TEST_KIND,
            "payload_format": PERFORMANCE_TEST_VERSION,
            "data": data,
            "checksum": hashlib.sha256(json.dumps(data).encode()).hexdigest(),
        }

        with httpx.Client() as client:
            response = client.post(self.results_url, json=payload)
            response.raise_for_status()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import NULL_VALUE, RelationshipStatus
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class AttributeAddQuery(Query):
    name = "attribute_add"
    type = QueryType.WRITE

    def __init__(
        self,
        node_kind: str,
        attribute_name: str,
        attribute_kind: str,
        branch_support: str,
        default_value: Any | None = None,
        **kwargs: Any,
    ) -> None:
        self.node_kind = node_kind
        self.attribute_name = attribute_name
        self.attribute_kind = attribute_kind
        self.branch_support = branch_support
        self.default_value = default_value

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["node_kind"] = self.node_kind
        self.params["attr_name"] = self.attribute_name
        self.params["attr_type"] = self.attribute_kind
        self.params["branch_support"] = self.branch_support
        self.params["current_time"] = self.at.to_string()

        if self.default_value is not None:
            self.params["attr_value"] = self.default_value
        else:
            self.params["attr_value"] = NULL_VALUE

        self.params["rel_props"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
        }

        self.params["is_protected_default"] = False
        self.params["is_visible_default"] = True

        query = """
        MERGE (av:AttributeValue { value: $attr_value, is_default: true })
        MERGE (is_protected_value:Boolean { value: $is_protected_default })
        MERGE (is_visible_value:Boolean { value: $is_visible_default })
        WITH av, is_protected_value, is_visible_value
        MATCH p = (n:%(node_kind)s)
        CALL {
            WITH n
            MATCH (root:Root)<-[r1:IS_PART_OF]-(n)
            OPTIONAL MATCH (n)-[r2:HAS_ATTRIBUTE]-(:Attribute { name: $attr_name })
            WHERE all(r in [r1, r2] WHERE (%(branch_filter)s))
            RETURN n as n1, r1 as r11, r2 as r12
            ORDER BY r2.branch_level DESC, r2.from ASC, r1.branch_level DESC, r1.from ASC
            LIMIT 1
        }
        WITH n1 as n, r11 as r1, r12 as r2, av, is_protected_value, is_visible_value
        WHERE r1.status = "active" AND (r2 IS NULL OR r2.status = "deleted")
        CREATE (a:Attribute { name: $attr_name, branch_support: $branch_support })
        CREATE (n)-[:HAS_ATTRIBUTE $rel_props ]->(a)
        CREATE (a)-[:HAS_VALUE $rel_props ]->(av)
        CREATE (a)-[:IS_PROTECTED $rel_props]->(is_protected_value)
        CREATE (a)-[:IS_VISIBLE $rel_props]->(is_visible_value)
        %(uuid_generation)s
        FOREACH (i in CASE WHEN r2.status = "deleted" THEN [1] ELSE [] END |
            SET r2.to = $current_time
        )
        """ % {
            "branch_filter": branch_filter,
            "node_kind": self.node_kind,
            "uuid_generation": db.render_uuid_generation(node_label="a", node_attr="uuid"),
        }
        self.add_to_query(query)
        self.return_labels = ["n.uuid", "a.uuid"]



from __future__ import annotations

from prefect import flow

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.merger.merger import DiffMerger
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.merge import BranchMerger
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.exceptions import ValidationError
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow
from infrahub.workflows.catalogue import BRANCH_MERGE
from infrahub.workflows.utils import add_tags


@flow(name="merge-branch-mutation", flow_run_name="Merge branch graphQL mutation")
async def merge_branch_mutation(branch: str, context: InfrahubContext, service: InfrahubServices) -> None:
    await add_tags(branches=[branch])

    async with service.database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=branch)
        base_branch = await Branch.get_by_name(db=db, name=registry.default_branch)

        component_registry = get_component_registry()
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
        enriched_diff_metadata = await diff_coordinator.update_branch_diff(base_branch=base_branch, diff_branch=obj)
        async for _ in diff_repository.get_all_conflicts_for_diff(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        ):
            # if there are any conflicts, raise the error
            raise ValidationError(
                f"Branch {obj.name} contains conflicts with the default branch."
                " Please create a Proposed Change to resolve the conflicts or manually update them before merging."
            )
        node_diff_field_summaries = await diff_repository.get_node_field_summaries(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        )

        merger = BranchMerger(
            db=db,
            diff_coordinator=diff_coordinator,
            diff_merger=diff_merger,
            diff_repository=diff_repository,
            source_branch=obj,
            service=service,
        )
        candidate_schema = merger.get_candidate_schema()
        determiner = ConstraintValidatorDeterminer(schema_branch=candidate_schema)
        constraints = await determiner.get_constraints(node_diffs=node_diff_field_summaries)
        if obj.has_schema_changes:
            constraints += await merger.calculate_validations(target_schema=candidate_schema)

        if constraints:
            responses = await schema_validate_migrations(
                message=SchemaValidateMigrationData(
                    branch=obj, schema_branch=candidate_schema, constraints=constraints
                ),
                service=service,
            )
            error_messages = [violation.message for response in responses for violation in response.violations]
            if error_messages:
                raise ValidationError(",\n".join(error_messages))

        await service.workflow.execute_workflow(workflow=BRANCH_MERGE, context=context, parameters={"branch": obj.name})



from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import APIRouter, Depends, Response

from infrahub import config, models
from infrahub.api.dependencies import get_access_token, get_db, get_refresh_token
from infrahub.auth import (
    AccountSession,
    authenticate_with_password,
    create_fresh_access_token,
    invalidate_refresh_token,
)

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

router = APIRouter(prefix="/auth")


@router.post("/login")
async def login_user(
    credentials: models.PasswordCredential, response: Response, db: InfrahubDatabase = Depends(get_db)
) -> models.UserToken:
    token = await authenticate_with_password(db=db, credentials=credentials)
    response.set_cookie(
        "access_token", token.access_token, httponly=True, max_age=config.SETTINGS.security.access_token_lifetime
    )
    response.set_cookie(
        "refresh_token", token.refresh_token, httponly=True, max_age=config.SETTINGS.security.refresh_token_lifetime
    )
    return token


@router.post("/refresh")
async def refresh_jwt_token(
    response: Response,
    db: InfrahubDatabase = Depends(get_db),
    refresh_token: models.RefreshTokenData = Depends(get_refresh_token),
) -> models.AccessTokenResponse:
    token = await create_fresh_access_token(db=db, refresh_data=refresh_token)
    response.set_cookie(
        "access_token", token.access_token, httponly=True, max_age=config.SETTINGS.security.access_token_lifetime
    )

    return token


@router.post("/logout")
async def logout(
    response: Response, db: InfrahubDatabase = Depends(get_db), user_session: AccountSession = Depends(get_access_token)
) -> None:
    if user_session.session_id:
        await invalidate_refresh_token(db=db, token_id=user_session.session_id)

    response.delete_cookie("access_token")
    response.delete_cookie("refresh_token")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

from typing_extensions import Self

if TYPE_CHECKING:
    from neo4j import AsyncResult, AsyncSession, AsyncTransaction, Record

    from infrahub.core.schema.schema_branch import SchemaBranch


@runtime_checkable
class Timestamp(Protocol): ...


@runtime_checkable
class NodeSchema(Protocol): ...


@runtime_checkable
class ProfileSchema(Protocol): ...


@runtime_checkable
class TemplateSchema(Protocol): ...


@runtime_checkable
class Branch(Protocol): ...


@runtime_checkable
class InfrahubDatabase(Protocol):
    @property
    def is_session(self) -> bool: ...

    @property
    def is_transaction(self) -> bool: ...

    def add_schema(self, schema: SchemaBranch, name: str | None = None) -> None: ...
    def start_session(self, read_only: bool = False, schemas: list[SchemaBranch] | None = None) -> Self: ...
    def start_transaction(self, schemas: list[SchemaBranch] | None = None) -> Self: ...
    async def session(self) -> AsyncSession: ...
    async def transaction(self, name: str | None) -> AsyncTransaction: ...
    async def close(self) -> None: ...

    async def execute_query(
        self, query: str, params: dict[str, Any] | None = None, name: str = "undefined"
    ) -> list[Record]: ...

    async def execute_query_with_metadata(
        self, query: str, params: dict[str, Any] | None = None, name: str = "undefined"
    ) -> tuple[list[Record], dict[str, Any]]: ...

    async def run_query(
        self, query: str, params: dict[str, Any] | None = None, name: str = "undefined"
    ) -> AsyncResult: ...

    def render_list_comprehension(self, items: str, item_name: str) -> str: ...
    def render_list_comprehension_with_list(self, items: str, item_names: list[str]) -> str: ...
    def render_uuid_generation(self, node_label: str, node_attr: str) -> str: ...
    def get_id_function_name(self) -> str: ...
    def to_database_id(self, db_id: str | int) -> str | int: ...


@runtime_checkable
class CoreNode(Protocol):
    id: str

    def get_id(self) -> str: ...
    def get_kind(self) -> str: ...
    @classmethod
    async def init(
        cls,
        schema: NodeSchema | ProfileSchema | str,
        db: InfrahubDatabase,
        branch: Branch | str | None = None,
        at: Timestamp | str | None = None,
    ) -> Self: ...
    async def new(self, db: InfrahubDatabase, id: str | None = None, **kwargs: Any) -> Self: ...
    async def save(self, db: InfrahubDatabase, at: Timestamp | None = None) -> Self: ...
    async def delete(self, db: InfrahubDatabase, at: Timestamp | None = None) -> None: ...
    async def load(
        self,
        db: InfrahubDatabase,
        id: str | None = None,
        db_id: str | None = None,
        updated_at: Timestamp | str | None = None,
        **kwargs: Any,
    ) -> Self: ...
    async def to_graphql(
        self,
        db: InfrahubDatabase,
        fields: dict | None = None,
        related_node_ids: set | None = None,
        filter_sensitive: bool = False,
        permissions: dict | None = None,
    ) -> dict: ...
    async def render_display_label(self, db: InfrahubDatabase | None = None) -> str: ...
    async def from_graphql(self, data: dict, db: InfrahubDatabase) -> bool: ...



from __future__ import annotations

from typing import Any

from .. import InfrahubClient, InfrahubClientSync
from ..config import Config
from ..ctl import config


def initialize_client(
    branch: str | None = None,
    identifier: str | None = None,
    timeout: int | None = None,
    max_concurrent_execution: int | None = None,
    retry_on_failure: bool | None = None,
) -> InfrahubClient:
    return InfrahubClient(
        config=_define_config(
            branch=branch,
            identifier=identifier,
            timeout=timeout,
            max_concurrent_execution=max_concurrent_execution,
            retry_on_failure=retry_on_failure,
        )
    )


def initialize_client_sync(
    branch: str | None = None,
    identifier: str | None = None,
    timeout: int | None = None,
    max_concurrent_execution: int | None = None,
    retry_on_failure: bool | None = None,
) -> InfrahubClientSync:
    return InfrahubClientSync(
        config=_define_config(
            branch=branch,
            identifier=identifier,
            timeout=timeout,
            max_concurrent_execution=max_concurrent_execution,
            retry_on_failure=retry_on_failure,
        )
    )


def _define_config(
    branch: str | None = None,
    identifier: str | None = None,
    timeout: int | None = None,
    max_concurrent_execution: int | None = None,
    retry_on_failure: bool | None = None,
) -> Config:
    client_config: dict[str, Any] = {
        "address": config.SETTINGS.active.server_address,
        "insert_tracker": True,
        "identifier": identifier,
    }

    if config.SETTINGS.active.api_token:
        client_config["api_token"] = config.SETTINGS.active.api_token

    if timeout:
        client_config["timeout"] = timeout

    if max_concurrent_execution is not None:
        client_config["max_concurrent_execution"] = max_concurrent_execution

    if retry_on_failure is not None:
        client_config["retry_on_failure"] = retry_on_failure

    if branch:
        client_config["default_branch"] = branch

    return Config(**client_config)



from infrahub.events.models import InfrahubEvent
from infrahub.utils import get_all_subclasses


def get_all_events() -> list[type[InfrahubEvent]]:
    """Recursively get all subclasses of the given class."""
    subclasses = get_all_subclasses(InfrahubEvent)
    return [cls for cls in subclasses if isinstance(cls.event_name, str)]


def get_all_infrahub_node_kind_events() -> list[str]:
    """Recursively get all events marked as infrahub.node.kind events.

    These events can be used to filter on a specific type for the webhooks
    """
    return [event.event_name for event in get_all_events() if event.infrahub_node_kind_event] + ["all"]



from infrahub.core.diff.calculator import DiffCalculator
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffCalculatorDependency(DependencyBuilder[DiffCalculator]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffCalculator:
        return DiffCalculator(db=context.db)



from infrahub.core.validators.attribute.choices import AttributeChoicesChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaAttributeChoicesConstraintDependency(DependencyBuilder[AttributeChoicesChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AttributeChoicesChecker:
        return AttributeChoicesChecker(db=context.db, branch=context.branch)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Literal, overload

from .exceptions import NodeInvalidError, NodeNotFoundError
from .node import parse_human_friendly_id

if TYPE_CHECKING:
    from .client import SchemaType, SchemaTypeSync
    from .node import InfrahubNode, InfrahubNodeSync
    from .protocols_base import CoreNode, CoreNodeSync


def get_schema_name(schema: type[SchemaType | SchemaTypeSync] | str | None = None) -> str | None:
    if isinstance(schema, str):
        return schema

    if hasattr(schema, "_is_runtime_protocol") and schema._is_runtime_protocol:  # type: ignore[union-attr]
        return schema.__name__  # type: ignore[union-attr]

    return None


class NodeStoreBranch:
    def __init__(self, name: str) -> None:
        self.branch_name = name

        self._objs: dict[str, InfrahubNode | InfrahubNodeSync | CoreNode | CoreNodeSync] = {}
        self._hfids: dict[str, dict[tuple, str]] = {}
        self._keys: dict[str, str] = {}
        self._uuids: dict[str, str] = {}

    def count(self) -> int:
        return len(self._objs)

    def set(self, node: InfrahubNode | InfrahubNodeSync | CoreNode | CoreNodeSync, key: str | None = None) -> None:
        self._objs[node._internal_id] = node

        if key:
            self._keys[key] = node._internal_id

        if node.id:
            self._uuids[node.id] = node._internal_id

        if hfid := node.get_human_friendly_id():
            for kind in node.get_all_kinds():
                if kind not in self._hfids:
                    self._hfids[kind] = {}
                self._hfids[kind][tuple(hfid)] = node._internal_id

    def get(
        self,
        key: str | list[str],
        kind: type[SchemaType | SchemaTypeSync] | str | None = None,
        raise_when_missing: bool = True,
    ) -> InfrahubNode | InfrahubNodeSync | CoreNode | CoreNodeSync | None:
        found_invalid = False

        kind_name = get_schema_name(schema=kind)

        if isinstance(key, list):
            try:
                return self._get_by_hfid(key, kind=kind_name)
            except NodeNotFoundError:
                pass

        elif isinstance(key, str):
            try:
                return self._get_by_internal_id(key, kind=kind_name)
            except NodeInvalidError:
                found_invalid = True
            except NodeNotFoundError:
                pass

            try:
                return self._get_by_id(key, kind=kind_name)
            except NodeInvalidError:
                found_invalid = True
            except NodeNotFoundError:
                pass

            try:
                return self._get_by_key(key, kind=kind_name)
            except NodeInvalidError:
                found_invalid = True
            except NodeNotFoundError:
                pass

            try:
                return self._get_by_hfid(key, kind=kind_name)
            except NodeNotFoundError:
                pass

        if not raise_when_missing:
            return None

        if kind and found_invalid:
            raise NodeInvalidError(
                identifier={"key": [key] if isinstance(key, str) else key},
                message=f"Found a node of a different kind instead of {kind} for key {key!r} in the store ({self.branch_name})",
            )

        raise NodeNotFoundError(
            identifier={"key": [key] if isinstance(key, str) else key},
            message=f"Unable to find the node {key!r} in the store ({self.branch_name})",
        )

    def _get_by_internal_id(
        self, internal_id: str, kind: str | None = None
    ) -> InfrahubNode | InfrahubNodeSync | CoreNode | CoreNodeSync:
        if internal_id not in self._objs:
            raise NodeNotFoundError(
                identifier={"internal_id": [internal_id]},
                message=f"Unable to find the node {internal_id!r} in the store ({self.branch_name})",
            )

        node = self._objs[internal_id]
        if kind and kind not in node.get_all_kinds():
            raise NodeInvalidError(
                node_type=kind,
                identifier={"internal_id": [internal_id]},
                message=f"Found a node of kind {node.get_kind()} instead of {kind} for internal_id {internal_id!r} in the store ({self.branch_name})",
            )

        return node

    def _get_by_key(
        self, key: str, kind: str | None = None
    ) -> InfrahubNode | InfrahubNodeSync | CoreNode | CoreNodeSync:
        if key not in self._keys:
            raise NodeNotFoundError(
                identifier={"key": [key]},
                message=f"Unable to find the node {key!r} in the store ({self.branch_name})",
            )

        node = self._get_by_internal_id(self._keys[key])

        if kind and node.get_kind() != kind:
            raise NodeInvalidError(
                node_type=kind,
                identifier={"key": [key]},
                message=f"Found a node of kind {node.get_kind()} instead of {kind} for key {key!r} in the store ({self.branch_name})",
            )

        return node

    def _get_by_id(self, id: str, kind: str | None = None) -> InfrahubNode | InfrahubNodeSync | CoreNode | CoreNodeSync:
        if id not in self._uuids:
            raise NodeNotFoundError(
                identifier={"id": [id]},
                message=f"Unable to find the node {id!r} in the store ({self.branch_name})",
            )

        node = self._get_by_internal_id(self._uuids[id])
        if kind and kind not in node.get_all_kinds():
            raise NodeInvalidError(
                node_type=kind,
                identifier={"id": [id]},
                message=f"Found a node of kind {node.get_kind()} instead of {kind} for id {id!r} in the store ({self.branch_name})",
            )

        return node

    def _get_by_hfid(
        self, hfid: str | list[str], kind: str | None = None
    ) -> InfrahubNode | InfrahubNodeSync | CoreNode | CoreNodeSync:
        if not kind:
            node_kind, node_hfid = parse_human_friendly_id(hfid)
        elif kind and isinstance(hfid, str) and hfid.startswith(kind):
            node_kind, node_hfid = parse_human_friendly_id(hfid)
        else:
            node_kind = kind
            node_hfid = [hfid] if isinstance(hfid, str) else hfid

        exception_to_raise_if_not_found = NodeNotFoundError(
            node_type=node_kind,
            identifier={"hfid": node_hfid},
            message=f"Unable to find the node {hfid!r} in the store ({self.branch_name})",
        )

        if node_kind not in self._hfids:
            raise exception_to_raise_if_not_found

        if tuple(node_hfid) not in self._hfids[node_kind]:
            raise exception_to_raise_if_not_found

        internal_id = self._hfids[node_kind][tuple(node_hfid)]
        return self._objs[internal_id]


class NodeStoreBase:
    """Internal Store for InfrahubNode objects.

    Often while creating a lot of new objects,
    we need to save them in order to reuse them later to associate them with another node for example.
    """

    def __init__(self, default_branch: str | None = None) -> None:
        self._branches: dict[str, NodeStoreBranch] = {}

        if default_branch is None:
            default_branch = "main"
            warnings.warn(
                "Using a store without specifying a default branch is deprecated and will be removed in a future version. "
                "Please explicitly specify a branch name.",
                DeprecationWarning,
                stacklevel=2,
            )

        self._default_branch = default_branch

    def _get_branch(self, branch: str | None = None) -> str:
        return branch or self._default_branch

    def _set(
        self,
        node: InfrahubNode | InfrahubNodeSync | SchemaType | SchemaTypeSync,
        key: str | None = None,
        branch: str | None = None,
    ) -> None:
        branch = self._get_branch(branch or node.get_branch())

        if branch not in self._branches:
            self._branches[branch] = NodeStoreBranch(name=branch)

        self._branches[branch].set(node=node, key=key)

    def _get(  # type: ignore[no-untyped-def]
        self,
        key: str | list[str],
        kind: type[SchemaType | SchemaTypeSync] | str | None = None,
        raise_when_missing: bool = True,
        branch: str | None = None,
    ):
        branch = self._get_branch(branch)

        if branch not in self._branches:
            self._branches[branch] = NodeStoreBranch(name=branch)

        return self._branches[branch].get(key=key, kind=kind, raise_when_missing=raise_when_missing)

    def count(self, branch: str | None = None) -> int:
        branch = self._get_branch(branch)

        if branch not in self._branches:
            return 0

        return self._branches[branch].count()


class NodeStore(NodeStoreBase):
    @overload
    def get(
        self,
        key: str | list[str],
        kind: type[SchemaType],
        raise_when_missing: Literal[True] = True,
        branch: str | None = ...,
    ) -> SchemaType: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: type[SchemaType],
        raise_when_missing: Literal[False] = False,
        branch: str | None = ...,
    ) -> SchemaType | None: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: type[SchemaType],
        raise_when_missing: bool = ...,
        branch: str | None = ...,
    ) -> SchemaType: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: str | None = ...,
        raise_when_missing: Literal[True] = True,
        branch: str | None = ...,
    ) -> InfrahubNode: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: str | None = ...,
        raise_when_missing: Literal[False] = False,
        branch: str | None = ...,
    ) -> InfrahubNode | None: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: str | None = ...,
        raise_when_missing: bool = ...,
        branch: str | None = ...,
    ) -> InfrahubNode: ...

    def get(
        self,
        key: str | list[str],
        kind: str | type[SchemaType] | None = None,
        raise_when_missing: bool = True,
        branch: str | None = None,
    ) -> InfrahubNode | SchemaType | None:
        return self._get(key=key, kind=kind, raise_when_missing=raise_when_missing, branch=branch)

    @overload
    def get_by_hfid(
        self, key: str | list[str], raise_when_missing: Literal[True] = True, branch: str | None = ...
    ) -> InfrahubNode: ...

    @overload
    def get_by_hfid(
        self, key: str | list[str], raise_when_missing: Literal[False] = False, branch: str | None = ...
    ) -> InfrahubNode | None: ...

    def get_by_hfid(
        self, key: str | list[str], raise_when_missing: bool = True, branch: str | None = None
    ) -> InfrahubNode | None:
        warnings.warn(
            "get_by_hfid() is deprecated and will be removed in a future version. Use get() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get(key=key, raise_when_missing=raise_when_missing, branch=branch)

    def set(self, node: InfrahubNode | SchemaType, key: str | None = None, branch: str | None = None) -> None:
        return self._set(node=node, key=key, branch=branch)


class NodeStoreSync(NodeStoreBase):
    @overload
    def get(
        self,
        key: str | list[str],
        kind: type[SchemaTypeSync],
        raise_when_missing: Literal[True] = True,
        branch: str | None = ...,
    ) -> SchemaTypeSync: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: type[SchemaTypeSync],
        raise_when_missing: Literal[False] = False,
        branch: str | None = ...,
    ) -> SchemaTypeSync | None: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: type[SchemaTypeSync],
        raise_when_missing: bool = ...,
        branch: str | None = ...,
    ) -> SchemaTypeSync: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: str | None = ...,
        raise_when_missing: Literal[True] = True,
        branch: str | None = ...,
    ) -> InfrahubNodeSync: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: str | None = ...,
        raise_when_missing: Literal[False] = False,
        branch: str | None = ...,
    ) -> InfrahubNodeSync | None: ...

    @overload
    def get(
        self,
        key: str | list[str],
        kind: str | None = ...,
        raise_when_missing: bool = ...,
        branch: str | None = ...,
    ) -> InfrahubNodeSync: ...

    def get(
        self,
        key: str | list[str],
        kind: str | type[SchemaTypeSync] | None = None,
        raise_when_missing: bool = True,
        branch: str | None = None,
    ) -> InfrahubNodeSync | SchemaTypeSync | None:
        return self._get(key=key, kind=kind, raise_when_missing=raise_when_missing, branch=branch)

    @overload
    def get_by_hfid(
        self, key: str | list[str], raise_when_missing: Literal[True] = True, branch: str | None = ...
    ) -> InfrahubNodeSync: ...

    @overload
    def get_by_hfid(
        self, key: str | list[str], raise_when_missing: Literal[False] = False, branch: str | None = ...
    ) -> InfrahubNodeSync | None: ...

    def get_by_hfid(
        self, key: str | list[str], raise_when_missing: bool = True, branch: str | None = None
    ) -> InfrahubNodeSync | None:
        warnings.warn(
            "get_by_hfid() is deprecated and will be removed in a future version. Use get() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get(key=key, raise_when_missing=raise_when_missing, branch=branch)

    def set(self, node: InfrahubNodeSync | SchemaTypeSync, key: str | None = None, branch: str | None = None) -> None:
        return self._set(node=node, key=key, branch=branch)



import asyncio

import typer
import ujson
from aio_pika.abc import AbstractIncomingMessage
from infrahub_sdk.async_typer import AsyncTyper
from rich import print as rprint

from infrahub import config
from infrahub.components import ComponentType
from infrahub.services import InfrahubServices
from infrahub.services.adapters.message_bus.rabbitmq import RabbitMQMessageBus

app = AsyncTyper()


async def print_event(event: AbstractIncomingMessage) -> None:
    message = {"routing_key": event.routing_key, "message": ujson.loads(event.body)}
    rprint(message)


@app.command()
async def listen(
    topic: str = "#", config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG")
) -> None:
    """Listen to event in the Events bus and print them."""
    config.load_and_exit(config_file)

    # Below code seems to be a duplicate of `RabbitMQMessageBus._initialize_api_server`.
    #  Also, here we are using ComponentType.NONE so RabbitMQMessageBus will NOT instantiate connection.
    #  We might want to factorize here and use ComponentType.GitAgent so broker instantiates connection correctly.
    component_type = ComponentType.NONE
    broker = await RabbitMQMessageBus.new(component_type=component_type)
    _ = await InfrahubServices.new(message_bus=broker, component_type=component_type)

    queue = await broker.channel.declare_queue(exclusive=True)
    exchange_name = f"{config.SETTINGS.broker.namespace}.events"
    exchange = await broker.channel.get_exchange(name=exchange_name)
    await queue.consume(print_event, no_ack=True)

    await queue.bind(exchange, routing_key=topic)
    print(f" Waiting for events matching the topic `{topic}`. To exit press CTRL+C")
    await asyncio.Future()



import hashlib
import json
import platform
import time
from typing import Any

from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.logging import get_run_logger

from infrahub import __version__, config
from infrahub.core import registry, utils
from infrahub.core.branch import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.services import InfrahubServices

from .constants import TELEMETRY_KIND, TELEMETRY_VERSION
from .database import gather_database_information
from .models import (
    TelemetryBranchData,
    TelemetryData,
    TelemetrySchemaData,
    TelemetryWorkerData,
)
from .task_manager import gather_prefect_information
from .utils import determine_infrahub_type


@task(name="telemetry-schema-information", task_run_name="Gather Schema Information", cache_policy=NONE)
async def gather_schema_information(branch: Branch) -> TelemetrySchemaData:
    main_schema = registry.schema.get_schema_branch(name=branch.name)
    return TelemetrySchemaData(
        node_count=len(main_schema.node_names),
        generic_count=len(main_schema.generic_names),
        last_update=branch.schema_changed_at or "",
    )


@task(name="telemetry-feature-information", task_run_name="Gather Feature Information", cache_policy=NONE)
async def gather_feature_information(service: InfrahubServices) -> dict[str, int]:
    async with service.database.start_session() as db:
        data = {}
        features_to_count = [
            InfrahubKind.ARTIFACT,
            InfrahubKind.RESOURCEPOOL,
            InfrahubKind.REPOSITORY,
            InfrahubKind.GENERICGROUP,
            InfrahubKind.PROFILE,
            InfrahubKind.PROPOSEDCHANGE,
            InfrahubKind.OBJECTTEMPLATE,
            InfrahubKind.TRANSFORM,
            InfrahubKind.WEBHOOK,
        ]
        for kind in features_to_count:
            data[kind] = await utils.count_nodes(db=db, label=kind)

        return data


@task(name="telemetry-gather-data", task_run_name="Gather Anonynous Data", cache_policy=NONE)
async def gather_anonymous_telemetry_data(service: InfrahubServices) -> TelemetryData:
    start_time = time.time()

    default_branch = registry.get_branch_from_registry()
    workers = await service.component.list_workers(branch=default_branch.name, schema_hash=False)

    data = TelemetryData(
        deployment_id=registry.id,
        execution_time=None,
        infrahub_version=__version__,
        infrahub_type=determine_infrahub_type(),
        python_version=platform.python_version(),
        platform=platform.machine(),
        workers=TelemetryWorkerData(
            total=len(workers),
            active=len([w for w in workers if w.active]),
        ),
        branches=TelemetryBranchData(
            total=len(registry.branch),
        ),
        features=await gather_feature_information(service=service),
        schema_info=await gather_schema_information(branch=default_branch),
        database=await gather_database_information(db=service.database),
        prefect=await gather_prefect_information(),
    )

    data.execution_time = time.time() - start_time

    return data


@task(name="telemetry-post-data", task_run_name="Upload data", retries=5, cache_policy=NONE)
async def post_telemetry_data(service: InfrahubServices, url: str, payload: dict[str, Any]) -> None:
    """Send the telemetry data to the specified URL, using HTTP POST."""
    response = await service.http.post(url=url, json=payload)
    response.raise_for_status()


@flow(name="anonymous_telemetry_send", flow_run_name="Send anonymous telemetry")
async def send_telemetry_push(service: InfrahubServices) -> None:
    log = get_run_logger()
    if config.SETTINGS.main.telemetry_optout:
        log.info("Skipping, User opted out of this service.")
        return

    log.info(f"Pushing anonymous telemetry data to {config.SETTINGS.main.telemetry_endpoint}...")

    data = await gather_anonymous_telemetry_data(service=service)
    data_dict = data.model_dump(mode="json")
    log.info(f"Anonymous usage telemetry gathered in {data.execution_time} seconds. | {data_dict}")

    payload = {
        "kind": TELEMETRY_KIND,
        "payload_format": TELEMETRY_VERSION,
        "data": data_dict,
        "checksum": hashlib.sha256(json.dumps(data_dict).encode()).hexdigest(),
    }

    await post_telemetry_data(service=service, url=config.SETTINGS.main.telemetry_endpoint, payload=payload)



from dataclasses import dataclass, field

from infrahub import config
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.diff.query_parser import DiffQueryParser
from infrahub.core.query.diff import (
    DiffCalculationQuery,
    DiffFieldPathsQuery,
    DiffNodePathsQuery,
    DiffPropertyPathsQuery,
)
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.log import get_logger

from .model.field_specifiers_map import NodeFieldSpecifierMap
from .model.path import CalculatedDiffs

log = get_logger()


@dataclass
class DiffCalculationRequest:
    base_branch: Branch
    diff_branch: Branch
    branch_from_time: Timestamp
    from_time: Timestamp
    to_time: Timestamp
    current_node_field_specifiers: NodeFieldSpecifierMap | None = field(default=None)
    new_node_field_specifiers: NodeFieldSpecifierMap | None = field(default=None)


class DiffCalculator:
    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db

    async def _run_diff_calculation_query(
        self,
        diff_parser: DiffQueryParser,
        query_class: type[DiffCalculationQuery],
        calculation_request: DiffCalculationRequest,
        limit: int,
    ) -> None:
        has_more_data = True
        offset = 0
        while has_more_data:
            diff_query = await query_class.init(
                db=self.db,
                branch=calculation_request.diff_branch,
                base_branch=calculation_request.base_branch,
                diff_branch_from_time=calculation_request.branch_from_time,
                diff_from=calculation_request.from_time,
                diff_to=calculation_request.to_time,
                current_node_field_specifiers=calculation_request.current_node_field_specifiers,
                new_node_field_specifiers=calculation_request.new_node_field_specifiers,
                limit=limit,
                offset=offset,
            )
            log.info(f"Beginning one diff calculation query {limit=}, {offset=}")
            await diff_query.execute(db=self.db)
            log.info("Diff calculation query complete")
            last_result = None
            for query_result in diff_query.get_results():
                diff_parser.read_result(query_result=query_result)
                last_result = query_result
            has_more_data = False
            if last_result:
                has_more_data = last_result.get_as_type("has_more_data", bool)
            offset += limit

    async def calculate_diff(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        from_time: Timestamp,
        to_time: Timestamp,
        include_unchanged: bool = True,
        previous_node_specifiers: NodeFieldSpecifierMap | None = None,
    ) -> CalculatedDiffs:
        if diff_branch.name == registry.default_branch:
            diff_branch_from_time = from_time
        else:
            diff_branch_from_time = Timestamp(diff_branch.get_branched_from())
        diff_parser = DiffQueryParser(
            base_branch=base_branch,
            diff_branch=diff_branch,
            schema_manager=registry.schema,
            from_time=from_time,
            to_time=to_time,
            previous_node_field_specifiers=previous_node_specifiers,
        )
        node_limit = int(config.SETTINGS.database.query_size_limit / 10)
        fields_limit = int(config.SETTINGS.database.query_size_limit / 3)
        properties_limit = int(config.SETTINGS.database.query_size_limit)

        calculation_request = DiffCalculationRequest(
            base_branch=base_branch,
            diff_branch=diff_branch,
            branch_from_time=diff_branch_from_time,
            from_time=from_time,
            to_time=to_time,
        )

        log.info("Beginning diff node-level calculation queries for branch")
        await self._run_diff_calculation_query(
            diff_parser=diff_parser,
            query_class=DiffNodePathsQuery,
            calculation_request=calculation_request,
            limit=node_limit,
        )
        log.info("Diff node-level calculation queries for branch complete")

        log.info("Beginning diff field-level calculation queries for branch")
        await self._run_diff_calculation_query(
            diff_parser=diff_parser,
            query_class=DiffFieldPathsQuery,
            calculation_request=calculation_request,
            limit=fields_limit,
        )
        log.info("Diff field-level calculation queries for branch complete")

        log.info("Beginning diff property-level calculation queries for branch")
        await self._run_diff_calculation_query(
            diff_parser=diff_parser,
            query_class=DiffPropertyPathsQuery,
            calculation_request=calculation_request,
            limit=properties_limit,
        )
        log.info("Diff property-level calculation queries for branch complete")

        if base_branch.name != diff_branch.name:
            current_node_field_specifiers = diff_parser.get_current_node_field_specifiers()
            new_node_field_specifiers = diff_parser.get_new_node_field_specifiers()
            calculation_request = DiffCalculationRequest(
                base_branch=base_branch,
                diff_branch=base_branch,
                branch_from_time=diff_branch_from_time,
                from_time=from_time,
                to_time=to_time,
                current_node_field_specifiers=current_node_field_specifiers,
                new_node_field_specifiers=new_node_field_specifiers,
            )

            log.info("Beginning diff node-level calculation queries for base")
            await self._run_diff_calculation_query(
                diff_parser=diff_parser,
                query_class=DiffNodePathsQuery,
                calculation_request=calculation_request,
                limit=node_limit,
            )
            log.info("Diff node-level calculation queries for base complete")

            log.info("Beginning diff field-level calculation queries for base")
            await self._run_diff_calculation_query(
                diff_parser=diff_parser,
                query_class=DiffFieldPathsQuery,
                calculation_request=calculation_request,
                limit=fields_limit,
            )
            log.info("Diff field-level calculation queries for base complete")

            log.info("Beginning diff property-level calculation queries for base")
            await self._run_diff_calculation_query(
                diff_parser=diff_parser,
                query_class=DiffPropertyPathsQuery,
                calculation_request=calculation_request,
                limit=properties_limit,
            )
            log.info("Diff property-level calculation queries for base complete")

        log.info("Parsing calculated diff")
        diff_parser.parse(include_unchanged=include_unchanged)
        log.info("Calculated diff parsed")
        return CalculatedDiffs(
            base_branch_name=base_branch.name,
            diff_branch_name=diff_branch.name,
            base_branch_diff=diff_parser.get_diff_root_for_branch(branch=base_branch.name),
            diff_branch_diff=diff_parser.get_diff_root_for_branch(branch=diff_branch.name),
        )



from typing import Any

from pydantic import BaseModel, Field

from infrahub.core.constants import PathType
from infrahub.core.schema import AttributeSchema, MainSchemaTypes, RelationshipSchema


class QueryRelationshipAttributePath(BaseModel):
    identifier: str
    attribute_name: str | None = Field(default=None)
    value: Any | None = Field(default=None)

    def __hash__(self) -> int:
        to_hash = self.identifier
        if self.attribute_name:
            to_hash += self.attribute_name
        if self.value:
            to_hash += str(self.value)
        return hash(to_hash)


class QueryAttributePath(BaseModel):
    attribute_name: str
    property_name: str | None = Field(default=None)
    value: Any | None = Field(default=None)

    def __hash__(self) -> int:
        to_hash = self.attribute_name
        if self.property_name:
            to_hash += self.property_name
        if self.value:
            to_hash += str(self.value)
        return hash(to_hash)


class NodeUniquenessQueryRequest(BaseModel):
    kind: str
    unique_attribute_paths: set[QueryAttributePath] = Field(default_factory=set)
    relationship_attribute_paths: set[QueryRelationshipAttributePath] = Field(default_factory=set)

    def __bool__(self) -> bool:
        if self.unique_attribute_paths or self.relationship_attribute_paths:
            return True
        return False

    def __str__(self) -> str:
        return (
            "ATTRS: "
            + "; ".join(
                q.attribute_name + " " + str(q.property_name) + " " + (str(q.value) if q.value is not None else "")
                for q in self.unique_attribute_paths
            )
            + " RELS: "
            + "; ".join(
                q.identifier + " " + str(q.attribute_name) + " " + (str(q.value) if q.value is not None else "")
                for q in self.relationship_attribute_paths
            )
        )


class NonUniqueRelatedAttribute(BaseModel):
    relationship: RelationshipSchema
    attribute_name: str
    attribute_value: str
    deepest_branch_name: str

    def __hash__(self) -> int:
        return hash(self.relationship.name + self.attribute_name + self.attribute_value)

    @property
    def grouping_key(self) -> str:
        return f"{self.relationship.name}/{self.attribute_name}/{self.attribute_value}"

    @property
    def path_type(self) -> PathType:
        return PathType.RELATIONSHIP_ONE

    @property
    def field_name(self) -> str:
        return self.relationship.name

    @property
    def property_name(self) -> str:
        return self.attribute_name


class NonUniqueAttribute(BaseModel):
    attribute: AttributeSchema
    attribute_name: str
    attribute_value: str
    deepest_branch_name: str

    def __hash__(self) -> int:
        return hash(self.attribute.name + self.attribute_name + self.attribute_value)

    @property
    def grouping_key(self) -> str:
        return f"{self.attribute_name}/{self.attribute_value}"

    @property
    def path_type(self) -> PathType:
        return PathType.ATTRIBUTE

    @property
    def field_name(self) -> str:
        return self.attribute_name

    @property
    def property_name(self) -> str:
        return "value"


class NonUniqueNode(BaseModel):
    node_schema: MainSchemaTypes
    node_id: str
    non_unique_attributes: list[NonUniqueAttribute] = Field(default_factory=list)
    non_unique_related_attributes: list[NonUniqueRelatedAttribute] = Field(default_factory=list)

    def get_relationship_violation(
        self, relationship_name: str, attribute_name: str | None
    ) -> NonUniqueRelatedAttribute | None:
        attribute_names = {attribute_name}
        if attribute_name is None:
            attribute_names.add("id")
        for nura in self.non_unique_related_attributes:
            if nura.relationship.name == relationship_name and nura.attribute_name in attribute_names:
                return nura
        return None

    def get_attribute_violation(self, attribute_name: str) -> NonUniqueAttribute | None:
        for nua in self.non_unique_attributes:
            if nua.attribute_name == attribute_name:
                return nua
        return None

    def get_constraint_violation(
        self, constraint_specifications: list[tuple[AttributeSchema | RelationshipSchema, str | None]]
    ) -> list[NonUniqueAttribute | NonUniqueRelatedAttribute] | None:
        violations: list[NonUniqueAttribute | NonUniqueRelatedAttribute] = []
        for sub_schema, property_name in constraint_specifications:
            if isinstance(sub_schema, AttributeSchema):
                attribute_violation = self.get_attribute_violation(sub_schema.name)
                if not attribute_violation:
                    return None
                violations.append(attribute_violation)
            elif isinstance(sub_schema, RelationshipSchema):
                relationship_violation = self.get_relationship_violation(sub_schema.name, property_name)
                if not relationship_violation:
                    return None
                violations.append(relationship_violation)
        return violations



from infrahub.core.diff.enricher.labels import DiffLabelsEnricher
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffLabelsEnricherDependency(DependencyBuilder[DiffLabelsEnricher]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffLabelsEnricher:
        return DiffLabelsEnricher(db=context.db)



from asyncio import run as aiorun
from datetime import datetime, timezone
from pathlib import Path

import typer
from rich.console import Console

from ..ctl.client import initialize_client
from ..transfer.exceptions import TransferError
from ..transfer.exporter.json import LineDelimitedJSONExporter
from .parameters import CONFIG_PARAM


def directory_name_with_timestamp() -> str:
    right_now = datetime.now(timezone.utc).astimezone()
    timestamp = right_now.strftime("%Y%m%d-%H%M%S")
    return f"infrahubexport-{timestamp}"


def dump(
    namespace: list[str] = typer.Option([], help="Namespace(s) to export"),
    directory: Path = typer.Option(directory_name_with_timestamp, help="Directory path to store export"),
    quiet: bool = typer.Option(False, help="No console output"),
    _: str = CONFIG_PARAM,
    branch: str = typer.Option(None, help="Branch from which to export"),
    concurrent: int = typer.Option(
        4,
        help="Maximum number of requests to execute at the same time.",
        envvar="INFRAHUB_MAX_CONCURRENT_EXECUTION",
    ),
    timeout: int = typer.Option(60, help="Timeout in sec", envvar="INFRAHUB_TIMEOUT"),
    exclude: list[str] = typer.Option(
        ["CoreAccount"],
        help="Prevent node kind(s) from being exported, CoreAccount is excluded by default",
    ),
) -> None:
    """Export nodes and their relationships out of the database."""
    console = Console()

    client = initialize_client(
        branch=branch, timeout=timeout, max_concurrent_execution=concurrent, retry_on_failure=True
    )

    exporter = LineDelimitedJSONExporter(client, console=Console() if not quiet else None)
    try:
        aiorun(exporter.export(export_directory=directory, namespaces=namespace, branch=branch, exclude=exclude))
    except TransferError as exc:
        console.print(f"[red]{exc}")
        raise typer.Exit(1)



from __future__ import annotations


class NodeFieldSpecifierMap:
    def __init__(self) -> None:
        # {uuid: {kind: {field_name, ...}}}
        self._map: dict[str, dict[str, set[str]]] = {}

    def __len__(self) -> int:
        return len(self._map)

    def __hash__(self) -> int:
        full_node_hash_sum = 0
        for node_uuid, node_dict in self._map.items():
            node_kinds_hash_sum = 0
            for kind, field_names in node_dict.items():
                fields_hash = hash(frozenset(field_names))
                node_kinds_hash_sum += hash(f"{hash(kind)}:{fields_hash}")
            full_node_hash_sum += hash(f"{node_uuid}:{node_kinds_hash_sum}")
        return hash(full_node_hash_sum)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NodeFieldSpecifierMap):
            return False
        return self._map == other._map

    def __sub__(self, other: NodeFieldSpecifierMap) -> NodeFieldSpecifierMap:
        subtracted = NodeFieldSpecifierMap()
        for node_uuid, node_dict in self._map.items():
            if node_uuid not in other._map:
                subtracted._map[node_uuid] = {**node_dict}
                continue
            subtracted_node_map = {}
            for kind, field_names in node_dict.items():
                subtracted_field_names = field_names - other._map[node_uuid].get(kind, set())
                if not subtracted_field_names:
                    continue
                subtracted_node_map[kind] = subtracted_field_names
            if not subtracted_node_map:
                continue
            subtracted._map[node_uuid] = subtracted_node_map
        return subtracted

    def add_entry(self, node_uuid: str, kind: str, field_name: str) -> None:
        if node_uuid not in self._map:
            self._map[node_uuid] = {}
        if kind not in self._map[node_uuid]:
            self._map[node_uuid][kind] = set()
        self._map[node_uuid][kind].add(field_name)

    def has_entry(self, node_uuid: str, kind: str, field_name: str) -> bool:
        return field_name in self._map.get(node_uuid, {}).get(kind, set())

    def get_uuids_list(self) -> list[str]:
        return list(self._map.keys())

    def get_uuid_field_names_map(self) -> dict[str, list[str]]:
        uuid_field_names_map: dict[str, list[str]] = {}
        for node_uuid, node_dict in self._map.items():
            field_names_set: set[str] = set()
            for field_names in node_dict.values():
                field_names_set |= field_names
            uuid_field_names_map[node_uuid] = list(field_names_set)
        return uuid_field_names_map



from typing import Any

from gunicorn.glogging import Logger

from infrahub.log import get_logger


class GunicornLogger(Logger):
    def __init__(self, cfg: Any) -> None:
        super().__init__(cfg)
        self.logger = get_logger("gunicorn")

    def critical(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.critical(msg, *args, **kwargs)

    def error(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.error(msg, *args, **kwargs)

    def warning(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.warning(msg, *args, **kwargs)

    def info(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.info(msg, *args, **kwargs)

    def debug(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.debug(msg, *args, **kwargs)

    def exception(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.exception(msg, *args, **kwargs)



import uuid
from typing import Any
from uuid import UUID

from prefect.client.orchestration import PrefectClient, get_client
from prefect.client.schemas.filters import (
    ArtifactFilter,
    ArtifactFilterType,
    FlowFilter,
    FlowFilterId,
    FlowFilterName,
    FlowRunFilter,
    FlowRunFilterId,
    FlowRunFilterName,
    FlowRunFilterState,
    FlowRunFilterStateType,
    FlowRunFilterTags,
    LogFilter,
    LogFilterFlowRunId,
)
from prefect.client.schemas.objects import Flow, FlowRun, StateType
from prefect.client.schemas.sorting import (
    FlowRunSort,
)

from infrahub.core.constants import TaskConclusion
from infrahub.core.query.node import NodeGetKindQuery
from infrahub.database import InfrahubDatabase
from infrahub.log import get_logger
from infrahub.utils import get_nested_dict
from infrahub.workflows.constants import TAG_NAMESPACE, WorkflowTag

from .constants import CONCLUSION_STATE_MAPPING
from .models import FlowLogs, FlowProgress, RelatedNodesInfo

log = get_logger()


class PrefectTask:
    @classmethod
    async def count_flow_runs(
        cls,
        client: PrefectClient,
        flow_filter: FlowFilter | None = None,
        flow_run_filter: FlowRunFilter | None = None,
    ) -> int:
        """
        Method to count the number of flow runs based on a flow_run_filter.
        The format of the body is the same as the one generated in read_flow_runs
        """
        body = {
            "flows": flow_filter.model_dump(mode="json") if flow_filter else None,
            "flow_runs": (flow_run_filter.model_dump(mode="json", exclude_unset=True) if flow_run_filter else None),
        }

        response = await client._client.post("/flow_runs/count", json=body)
        response.raise_for_status()
        return response.json()

    @classmethod
    async def _get_related_nodes(cls, db: InfrahubDatabase, flows: list[FlowRun]) -> RelatedNodesInfo:
        related_nodes = RelatedNodesInfo()

        # Extract all related nodes ID from tags
        for flow in flows:
            related_node_tag_prefix = WorkflowTag.RELATED_NODE.render(identifier="")
            related_node_ids = [
                tag.replace(related_node_tag_prefix, "") for tag in flow.tags if tag.startswith(related_node_tag_prefix)
            ]
            if not related_node_ids:
                continue
            related_nodes.add_nodes(flow_id=flow.id, node_ids=related_node_ids)

        if unique_related_node_ids := related_nodes.get_unique_related_node_ids():
            query = await NodeGetKindQuery.init(db=db, ids=unique_related_node_ids)
            await query.execute(db=db)
            unique_related_node_ids_kind = await query.get_node_kind_map()

            for node_id, node_kind in unique_related_node_ids_kind.items():
                if node_id in related_nodes.nodes:
                    related_nodes.nodes[node_id].kind = node_kind

        return related_nodes

    @classmethod
    async def _get_logs(cls, client: PrefectClient, flow_ids: list[UUID]) -> FlowLogs:
        logs_flow = FlowLogs()
        all_logs = await client.read_logs(log_filter=LogFilter(flow_run_id=LogFilterFlowRunId(any_=flow_ids)))
        for flow_log in all_logs:
            if flow_log.flow_run_id and flow_log.message not in ["Finished in state Completed()"]:
                logs_flow.logs[flow_log.flow_run_id].append(flow_log)

        return logs_flow

    @classmethod
    async def _get_flows(
        cls, client: PrefectClient, ids: list[UUID] | None = None, names: list[str] | None = None
    ) -> list[Flow]:
        if not names and not ids:
            return await client.read_flows()

        flow_filter = FlowFilter()
        flow_filter.name = FlowFilterName(any_=names) if names else None
        flow_filter.id = FlowFilterId(any_=ids) if ids else None
        return await client.read_flows(flow_filter=flow_filter)

    @classmethod
    async def _get_progress(cls, client: PrefectClient, flow_ids: list[UUID]) -> FlowProgress:
        artifacts = await client.read_artifacts(
            artifact_filter=ArtifactFilter(type=ArtifactFilterType(any_=["progress"])),
            flow_run_filter=FlowRunFilter(id=FlowRunFilterId(any_=flow_ids)),
        )
        flow_progress = FlowProgress()
        for artifact in artifacts:
            if artifact.flow_run_id in flow_progress.data:
                log.warning(
                    f"Multiple Progress Artifact found for the flow_run {artifact.flow_run_id}, keeping the first one"
                )
                continue
            if artifact.flow_run_id and isinstance(artifact.data, float):
                flow_progress.data[artifact.flow_run_id] = artifact.data

        return flow_progress

    @classmethod
    async def _extract_branch_name(cls, flow: FlowRun) -> str | None:
        branch_name = [
            tag.replace(WorkflowTag.BRANCH.render(identifier=""), "")
            for tag in flow.tags
            if tag.startswith(WorkflowTag.BRANCH.render(identifier=""))
        ]

        return branch_name[0] if branch_name else None

    @classmethod
    def _generate_flow_filter(cls, workflows: list[str] | None = None) -> FlowFilter:
        flow_filter = FlowFilter()
        if workflows:
            flow_filter.name = FlowFilterName(any_=workflows)
        return flow_filter

    @classmethod
    def _generate_flow_run_filter(
        cls,
        q: str | None = None,
        ids: list[str] | None = None,
        related_nodes: list[str] | None = None,
        statuses: list[StateType] | None = None,
        tags: list[str] | None = None,
        branch: str | None = None,
    ) -> FlowRunFilter:
        filter_tags = [TAG_NAMESPACE]

        if tags:
            filter_tags.extend(tags)
        if branch:
            filter_tags.append(WorkflowTag.BRANCH.render(identifier=branch))
        # We only support one related node for now, need to investigate HOW (and IF) we can support more

        if related_nodes:
            filter_tags.append(WorkflowTag.RELATED_NODE.render(identifier=related_nodes[0]))

        flow_run_filter = FlowRunFilter(
            tags=FlowRunFilterTags(all_=filter_tags),
        )
        if ids:
            flow_run_filter.id = FlowRunFilterId(any_=[uuid.UUID(id) for id in ids])

        if statuses:
            flow_run_filter.state = FlowRunFilterState(type=FlowRunFilterStateType(any_=statuses))

        if q:
            flow_run_filter.name = FlowRunFilterName(like_=q)

        return flow_run_filter

    @classmethod
    async def query(
        cls,
        db: InfrahubDatabase,
        fields: dict[str, Any],
        q: str | None = None,
        ids: list[str] | None = None,
        related_nodes: list[str] | None = None,
        statuses: list[StateType] | None = None,
        workflows: list[str] | None = None,
        tags: list[str] | None = None,
        branch: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        nodes: list[dict] = []
        count = None

        node_fields = get_nested_dict(nested_dict=fields, keys=["edges", "node"])
        log_fields = get_nested_dict(nested_dict=fields, keys=["edges", "node", "logs", "edges", "node"])
        logs_flow = FlowLogs()
        progress_flow = FlowProgress()
        workflow_names: dict[UUID, str] = {}
        related_nodes_info = RelatedNodesInfo()

        async with get_client(sync_client=False) as client:
            flow_filter = cls._generate_flow_filter(workflows=workflows)
            flow_run_filter = cls._generate_flow_run_filter(
                q=q, ids=ids, related_nodes=related_nodes, statuses=statuses, tags=tags, branch=branch
            )

            if "count" in fields:
                count = await cls.count_flow_runs(
                    client=client, flow_filter=flow_filter, flow_run_filter=flow_run_filter
                )

            if node_fields:
                flows = await client.read_flow_runs(
                    flow_filter=flow_filter,
                    flow_run_filter=flow_run_filter,
                    limit=limit,
                    offset=offset or 0,
                    sort=FlowRunSort.START_TIME_DESC,
                )
                if log_fields:
                    logs_flow = await cls._get_logs(client=client, flow_ids=[flow.id for flow in flows])

                if "progress" in node_fields:
                    progress_flow = await cls._get_progress(client=client, flow_ids=[flow.id for flow in flows])

                if (
                    "related_nodes" in node_fields
                    or "related_node" in node_fields
                    or "related_node_kind" in node_fields
                ):
                    related_nodes_info = await cls._get_related_nodes(db=db, flows=flows)

                if "workflow" in node_fields:
                    unique_flow_ids = {flow.flow_id for flow in flows}
                    workflow_names = {
                        flow.id: flow.name for flow in await cls._get_flows(client=client, ids=list(unique_flow_ids))
                    }

                for flow in flows:
                    logs = []

                    if log_fields:
                        logs = logs_flow.to_graphql(flow_id=flow.id)

                    related_node = related_nodes_info.get_first_related_node(flow_id=flow.id)

                    nodes.append(
                        {
                            "node": {
                                "title": flow.name,
                                "conclusion": CONCLUSION_STATE_MAPPING.get(
                                    str(flow.state_name), TaskConclusion.UNKNOWN
                                ).value,
                                "state": flow.state_type,
                                "progress": progress_flow.data.get(flow.id, None),
                                "parameters": flow.parameters,
                                "branch": await cls._extract_branch_name(flow=flow),
                                "tags": flow.tags,
                                "workflow": workflow_names.get(flow.flow_id, None),
                                "related_node": related_node.id if related_node else None,
                                "related_node_kind": related_node.kind if related_node else None,
                                "related_nodes": related_nodes_info.get_related_nodes_as_dict(flow_id=flow.id),
                                "created_at": flow.created.to_iso8601_string(),  # type: ignore
                                "updated_at": flow.updated.to_iso8601_string(),  # type: ignore
                                "start_time": flow.start_time.to_iso8601_string() if flow.start_time else None,
                                "id": flow.id,
                                "logs": {"edges": logs},
                            }
                        }
                    )

        return {"count": count or 0, "edges": nodes}



from infrahub_sdk.protocols import CoreGeneratorValidator
from infrahub_sdk.uuidt import UUIDT
from prefect import flow
from prefect.logging import get_run_logger

from infrahub.core.constants import InfrahubKind
from infrahub.core.timestamp import Timestamp
from infrahub.message_bus import InfrahubMessage, Meta, messages
from infrahub.message_bus.types import KVTTL
from infrahub.services import InfrahubServices
from infrahub.validators.tasks import start_validator
from infrahub.workflows.utils import add_tags


@flow(
    name="generator-definition-check",
    flow_run_name="Validate Generator selection for {message.generator_definition.definition_name}",
)
async def check(message: messages.RequestGeneratorDefinitionCheck, service: InfrahubServices) -> None:
    log = get_run_logger()
    await add_tags(branches=[message.source_branch], nodes=[message.proposed_change])
    events: list[InfrahubMessage] = []

    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=message.proposed_change)

    validator_name = f"Generator Validator: {message.generator_definition.definition_name}"
    validator_execution_id = str(UUIDT())
    check_execution_ids: list[str] = []

    await proposed_change.validations.fetch()

    previous_validator: CoreGeneratorValidator | None = None
    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer
        if (
            existing_validator.typename == InfrahubKind.GENERATORVALIDATOR
            and existing_validator.definition.id == message.generator_definition.definition_id
        ):
            previous_validator = existing_validator

    validator = await start_validator(
        service=service,
        validator=previous_validator,
        validator_type=CoreGeneratorValidator,
        proposed_change=message.proposed_change,
        data={
            "label": validator_name,
            "definition": message.generator_definition.definition_id,
        },
        context=message.context,
    )

    group = await service.client.get(
        kind=InfrahubKind.GENERICGROUP,
        prefetch_relationships=True,
        populate_store=True,
        id=message.generator_definition.group_id,
        branch=message.source_branch,
    )
    await group.members.fetch()

    existing_instances = await service.client.filters(
        kind=InfrahubKind.GENERATORINSTANCE,
        definition__ids=[message.generator_definition.definition_id],
        include=["object"],
        branch=message.source_branch,
    )
    instance_by_member = {}
    for instance in existing_instances:
        instance_by_member[instance.object.peer.id] = instance.id

    repository = message.branch_diff.get_repository(repository_id=message.generator_definition.repository_id)
    requested_instances = 0
    impacted_instances = message.branch_diff.get_subscribers_ids(kind=InfrahubKind.GENERATORINSTANCE)

    for relationship in group.members.peers:
        member = relationship.peer
        generator_instance = instance_by_member.get(member.id)
        if _run_generator(
            instance_id=generator_instance,
            managed_branch=message.source_branch_sync_with_git,
            impacted_instances=impacted_instances,
        ):
            check_execution_id = str(UUIDT())
            check_execution_ids.append(check_execution_id)
            requested_instances += 1
            log.info(f"Trigger execution of {message.generator_definition.definition_name} for {member.display_label}")
            events.append(
                messages.CheckGeneratorRun(
                    context=message.context,
                    generator_definition=message.generator_definition,
                    generator_instance=generator_instance,
                    commit=repository.source_commit,
                    repository_id=repository.repository_id,
                    repository_name=repository.repository_name,
                    repository_kind=repository.kind,
                    branch_name=message.source_branch,
                    query=message.generator_definition.query_name,
                    variables=member.extract(params=message.generator_definition.parameters),
                    target_id=member.id,
                    target_name=member.display_label,
                    validator_id=validator.id,
                    proposed_change=message.proposed_change,
                    meta=Meta(validator_execution_id=validator_execution_id, check_execution_id=check_execution_id),
                )
            )

    checks_in_execution = ",".join(check_execution_ids)
    await service.cache.set(
        key=f"validator_execution_id:{validator_execution_id}:checks",
        value=checks_in_execution,
        expires=KVTTL.TWO_HOURS,
    )
    events.append(
        messages.FinalizeValidatorExecution(
            start_time=Timestamp().to_string(),
            validator_id=validator.id,
            validator_execution_id=validator_execution_id,
            validator_type=InfrahubKind.GENERATORVALIDATOR,
            context=message.context,
            proposed_change=message.proposed_change,
        )
    )
    for event in events:
        event.assign_meta(parent=message)
        await service.message_bus.send(message=event)


def _run_generator(instance_id: str | None, managed_branch: bool, impacted_instances: list[str]) -> bool:
    """Returns a boolean to indicate if a generator instance needs to be executed
    Will return true if:
        * The instance_id wasn't set which could be that it's a new object that doesn't have a previous generator instance
        * The source branch is set to sync with Git which would indicate that it could contain updates in git to the generator
        * The instance_id exists in the impacted_instances list
    Will return false if:
        * The source branch is a not one that syncs with git and the instance_id exists and is not in the impacted list
    """
    if not instance_id or managed_branch:
        return True
    return instance_id in impacted_instances



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.core.node.standard import StandardNode
    from infrahub.database import InfrahubDatabase


class StandardNodeQuery(Query):
    def __init__(
        self, node: StandardNode = None, node_id: str | None = None, node_db_id: int | None = None, **kwargs: Any
    ):
        self.node = node
        self.node_id = node_id
        self.node_db_id = node_db_id

        if not self.node_id and self.node:
            self.node_id = self.node.uuid

        if not self.node_db_id and self.node:
            self.node_db_id = self.node.id

        super().__init__(**kwargs)


class RootNodeCreateQuery(StandardNodeQuery):
    name = "standard_node_create"
    type = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        node_type = self.node.get_type()
        self.params["node_prop"] = self.node.to_db()

        query = """
        CREATE (n:%s $node_prop)
        """ % (node_type)

        self.add_to_query(query=query)
        self.return_labels = ["n"]


class StandardNodeCreateQuery(StandardNodeQuery):
    name = "standard_node_create"
    type = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        node_type = self.node.get_type()
        self.params["node_prop"] = self.node.to_db()

        query = """
        MATCH (root:Root)
        CREATE (n:%s $node_prop)-[r:IS_PART_OF]->(root)
        """ % (node_type)

        self.add_to_query(query=query)
        self.return_labels = ["n"]


class StandardNodeUpdateQuery(StandardNodeQuery):
    name = "standard_node_update"
    type = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.node.get_type()
        self.params["node_prop"] = self.node.to_db()
        self.params["node_prop"]["uuid"] = str(self.node.uuid)
        self.params["uuid"] = str(self.node.uuid)

        query = """
        MATCH (n:%s { uuid: $uuid })
        SET n = $node_prop
        """ % (self.node.get_type(),)

        self.add_to_query(query=query)
        self.return_labels = ["n"]


class StandardNodeDeleteQuery(StandardNodeQuery):
    name = "standard_node_delete"
    insert_return = False
    type = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        query = """
        MATCH (n:%s { uuid: $uuid })
        DETACH DELETE (n)
        """ % (self.node.get_type())

        self.params["uuid"] = str(self.node_id)
        self.add_to_query(query)


class StandardNodeGetItemQuery(Query):
    name = "standard_node_get"
    type = QueryType.READ

    def __init__(self, node_id: str, node_type: str, **kwargs: Any) -> None:
        self.node_id = node_id
        self.node_type = node_type

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        query = """
            MATCH (n:%(node_type)s)
            WHERE %(id_func)s(n) = $node_id OR n.uuid = $node_id
            """ % {"node_type": self.node_type, "id_func": db.get_id_function_name()}

        self.params["node_id"] = self.node_id
        self.add_to_query(query)

        self.return_labels = ["n"]


class StandardNodeGetListQuery(Query):
    name = "standard_node_list"
    type = QueryType.READ

    def __init__(
        self, node_class: StandardNode, ids: list[str] | None = None, node_name: str | None = None, **kwargs: Any
    ) -> None:
        self.ids = ids
        self.node_name = node_name
        self.node_class = node_class

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        filters = []
        if self.ids:
            filters.append("n.uuid in $ids_value")
            self.params["ids_value"] = self.ids
        if self.node_name:
            filters.append("n.name = $name")
            self.params["name"] = self.node_name

        where = ""
        if filters:
            where = f"WHERE {' AND '.join(filters)}"

        query = """
        MATCH (n:%s)
        %s
        """ % (
            self.node_class.get_type(),
            where,
        )

        self.add_to_query(query)

        self.return_labels = ["n"]
        self.order_by = [f"{db.get_id_function_name()}(n)"]



from typing import ClassVar

from pydantic import Field

from .constants import EVENT_NAMESPACE
from .models import InfrahubEvent


class CommitUpdatedEvent(InfrahubEvent):
    """Event generated when the the commit within a repository has been updated."""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.repository.update_commit"

    commit: str = Field(..., description="The commit the repository was updated to")
    repository_id: str = Field(..., description="The ID of the repository")
    repository_name: str = Field(..., description="The name of the repository")

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.repository.{self.repository_id}",
            "infrahub.repository.name": self.repository_name,
            "infrahub.repository.id": self.repository_id,
            "infrahub.branch.name": self.meta.context.branch.name,
        }



from infrahub_sdk import InfrahubClient
from infrahub_sdk.node import InfrahubNode


async def group_add_subscriber(
    client: InfrahubClient, group: InfrahubNode, subscribers: list[str], branch: str
) -> dict:
    subscribers_str = ["{ id: " + f'"{subscriber}"' + " }" for subscriber in subscribers]
    query = """
    mutation {
        RelationshipAdd(
            data: {
                id: "%s",
                name: "subscribers",
                nodes: [ %s ]
            }
        ) {
            ok
        }
    }
    """ % (
        group.id,
        ", ".join(subscribers_str),
    )

    return await client.execute_graphql(query=query, branch_name=branch, tracker="mutation-relationshipadd")



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.diff.model.path import BranchTrackingId
from infrahub.core.diff.query.merge import DiffMergePropertiesQuery, DiffMergeQuery, DiffMergeRollbackQuery
from infrahub.log import get_logger

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.diff.model.path import EnrichedDiffRoot
    from infrahub.core.diff.repository.repository import DiffRepository
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase

    from .serializer import DiffMergeSerializer

log = get_logger()


class DiffMerger:
    def __init__(
        self,
        db: InfrahubDatabase,
        source_branch: Branch,
        destination_branch: Branch,
        diff_repository: DiffRepository,
        serializer: DiffMergeSerializer,
    ):
        self.source_branch = source_branch
        self.destination_branch = destination_branch
        self.db = db
        self.diff_repository = diff_repository
        self.serializer = serializer

    async def merge_graph(self, at: Timestamp) -> EnrichedDiffRoot:
        tracking_id = BranchTrackingId(name=self.source_branch.name)
        enriched_diffs = await self.diff_repository.get_roots_metadata(
            diff_branch_names=[self.source_branch.name],
            base_branch_names=[self.destination_branch.name],
            tracking_id=tracking_id,
        )
        latest_diff = None
        for diff in enriched_diffs:
            if latest_diff is None or (diff.to_time > latest_diff.to_time):
                latest_diff = diff
        if latest_diff is None:
            raise RuntimeError(f"Missing diff for branch {self.source_branch.name}")
        log.info(f"Retrieving diff {latest_diff.uuid}")
        enriched_diff = await self.diff_repository.get_one(
            diff_branch_name=self.source_branch.name, diff_id=latest_diff.uuid
        )
        log.info(f"Diff {latest_diff.uuid} retrieved")
        batch_num = 0
        async for node_diff_dicts, property_diff_dicts in self.serializer.serialize_diff(diff=enriched_diff):
            if node_diff_dicts:
                log.info(f"Merging batch of nodes #{batch_num}")
                merge_query = await DiffMergeQuery.init(
                    db=self.db,
                    branch=self.source_branch,
                    at=at,
                    target_branch=self.destination_branch,
                    node_diff_dicts=node_diff_dicts,
                )
                await merge_query.execute(db=self.db)
            if property_diff_dicts:
                log.info(f"Merging batch of properties #{batch_num}")
                merge_properties_query = await DiffMergePropertiesQuery.init(
                    db=self.db,
                    branch=self.source_branch,
                    at=at,
                    target_branch=self.destination_branch,
                    property_diff_dicts=property_diff_dicts,
                )
                await merge_properties_query.execute(db=self.db)
            log.info(f"Batch #{batch_num} merged")
            batch_num += 1

        self.source_branch.branched_from = at.to_string()
        await self.source_branch.save(db=self.db)
        registry.branch[self.source_branch.name] = self.source_branch
        return enriched_diff

    async def rollback(self, at: Timestamp) -> None:
        rollback_query = await DiffMergeRollbackQuery.init(
            db=self.db, branch=self.source_branch, target_branch=self.destination_branch, at=at
        )
        await rollback_query.execute(db=self.db)



import logging

import typer
from rich.console import Console
from rich.table import Table

from ..async_typer import AsyncTyper
from ..utils import calculate_time_diff
from .client import initialize_client
from .parameters import CONFIG_PARAM
from .utils import catch_exception

app = AsyncTyper()
console = Console()


DEFAULT_CONFIG_FILE = "infrahubctl.toml"
ENVVAR_CONFIG_FILE = "INFRAHUBCTL_CONFIG"


@app.callback()
def callback() -> None:
    """
    Manage the branches in a remote Infrahub instance.

    List, create, merge, rebase ..
    """


@app.command("list")
@catch_exception(console=console)
async def list_branch(_: str = CONFIG_PARAM) -> None:
    """List all existing branches."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    branches = await client.branch.all()

    table = Table(title="List of all branches")

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Description")
    table.add_column("Origin Branch")
    table.add_column("Branched From")
    table.add_column("Sync with Git")
    table.add_column("Has Schema Changes")
    table.add_column("Is Default")

    # identify the default branch and always print it first
    default_branch = [branch for branch in branches.values() if branch.is_default][0]
    table.add_row(
        default_branch.name,
        default_branch.description or " - ",
        default_branch.origin_branch,
        f"{default_branch.branched_from} ({calculate_time_diff(default_branch.branched_from)})",
        "[green]True" if default_branch.sync_with_git else "[#FF7F50]False",
        "[green]True" if default_branch.has_schema_changes else "[#FF7F50]False",
        "[green]True" if default_branch.is_default else "[#FF7F50]False",
    )

    for branch in branches.values():
        if branch.is_default:
            continue

        table.add_row(
            branch.name,
            branch.description or " - ",
            branch.origin_branch,
            f"{branch.branched_from} ({calculate_time_diff(branch.branched_from)})",
            "[green]True" if branch.sync_with_git else "[#FF7F50]False",
            "[green]True" if default_branch.has_schema_changes else "[#FF7F50]False",
            "[green]True" if branch.is_default else "[#FF7F50]False",
        )

    console.print(table)


@app.command()
@catch_exception(console=console)
async def create(
    branch_name: str = typer.Argument(..., help="Name of the branch to create"),
    description: str = typer.Option(default="", help="Description of the branch"),
    sync_with_git: bool = typer.Option(
        False, help="Extend the branch to Git and have Infrahub create the branch in connected repositories."
    ),
    isolated: bool = typer.Option(True, hidden=True, help="Set the branch to isolated mode (deprecated)"),  # noqa: ARG001
    _: str = CONFIG_PARAM,
) -> None:
    """Create a new branch."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    branch = await client.branch.create(branch_name=branch_name, description=description, sync_with_git=sync_with_git)
    console.print(f"Branch {branch_name!r} created successfully ({branch.id}).")


@app.command()
@catch_exception(console=console)
async def delete(branch_name: str, _: str = CONFIG_PARAM) -> None:
    """Delete a branch."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    await client.branch.delete(branch_name=branch_name)
    console.print(f"Branch '{branch_name}' deleted successfully.")


@app.command()
@catch_exception(console=console)
async def rebase(branch_name: str, _: str = CONFIG_PARAM) -> None:
    """Rebase a Branch with main."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    await client.branch.rebase(branch_name=branch_name)
    console.print(f"Branch '{branch_name}' rebased successfully.")


@app.command()
@catch_exception(console=console)
async def merge(branch_name: str, _: str = CONFIG_PARAM) -> None:
    """Merge a Branch with main."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    await client.branch.merge(branch_name=branch_name)
    console.print(f"Branch '{branch_name}' merged successfully.")


@app.command()
@catch_exception(console=console)
async def validate(branch_name: str, _: str = CONFIG_PARAM) -> None:
    """Validate if a branch has some conflict and is passing all the tests (NOT IMPLEMENTED YET)."""

    client = initialize_client()
    await client.branch.validate(branch_name=branch_name)
    console.print(f"Branch '{branch_name}' is valid.")



from __future__ import annotations

import os
import ssl
import sys
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any

import toml
from infrahub_sdk.utils import generate_uuid
from pydantic import (
    AliasChoices,
    BaseModel,
    Field,
    PrivateAttr,
    ValidationError,
    computed_field,
    field_validator,
    model_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self

from infrahub.constants.database import DatabaseType
from infrahub.exceptions import InitializationError, ProcessingError

if TYPE_CHECKING:
    from infrahub.services.adapters.cache import InfrahubCache
    from infrahub.services.adapters.message_bus import InfrahubMessageBus
    from infrahub.services.adapters.workflow import InfrahubWorkflow


VALID_DATABASE_NAME_REGEX = r"^[a-z][a-z0-9\.]+$"
THIRTY_DAYS_IN_SECONDS = 3600 * 24 * 30


def default_cors_allow_methods() -> list[str]:
    return ["DELETE", "GET", "OPTIONS", "PATCH", "POST", "PUT"]


def default_cors_allow_headers() -> list[str]:
    return ["accept", "authorization", "content-type", "user-agent", "x-csrftoken", "x-requested-with"]


def default_append_git_suffix_domains() -> list[str]:
    return ["github.com", "gitlab.com"]


class UserInfoMethod(str, Enum):
    POST = "post"
    GET = "get"


class SSOProtocol(str, Enum):
    OAUTH2 = "oauth2"
    OIDC = "oidc"


class Oauth2Provider(str, Enum):
    GOOGLE = "google"
    PROVIDER1 = "provider1"
    PROVIDER2 = "provider2"


class OIDCProvider(str, Enum):
    GOOGLE = "google"
    PROVIDER1 = "provider1"
    PROVIDER2 = "provider2"


class SSOInfo(BaseModel):
    providers: list[SSOProviderInfo] = Field(default_factory=list)

    @computed_field
    def enabled(self) -> bool:
        return bool(self.providers)


class SSOProviderInfo(BaseModel):
    name: str
    display_label: str
    icon: str
    protocol: SSOProtocol

    @computed_field
    def authorize_path(self) -> str:
        return f"/api/{self.protocol.value}/{self.name}/authorize"

    @computed_field
    def token_path(self) -> str:
        return f"/api/{self.protocol.value}/{self.name}/token"


class StorageDriver(str, Enum):
    FileSystemStorage = "local"
    InfrahubS3ObjectStorage = "s3"


class TraceExporterType(str, Enum):
    CONSOLE = "console"
    OTLP = "otlp"
    # JAEGER = "jaeger"
    # ZIPKIN = "zipkin"


class TraceTransportProtocol(str, Enum):
    GRPC = "grpc"
    HTTP_PROTOBUF = "http/protobuf"
    # HTTP_JSON = "http/json"


class BrokerDriver(str, Enum):
    RabbitMQ = "rabbitmq"
    NATS = "nats"

    @property
    def driver_module_path(self) -> str:
        match self:
            case BrokerDriver.NATS:
                return "infrahub.services.adapters.message_bus.nats"
            case BrokerDriver.RabbitMQ:
                return "infrahub.services.adapters.message_bus.rabbitmq"

    @property
    def driver_class_name(self) -> str:
        match self:
            case BrokerDriver.NATS:
                return "NATSMessageBus"
            case BrokerDriver.RabbitMQ:
                return "RabbitMQMessageBus"


class CacheDriver(str, Enum):
    Redis = "redis"
    NATS = "nats"

    @property
    def driver_module_path(self) -> str:
        match self:
            case CacheDriver.NATS:
                return "infrahub.services.adapters.cache.nats"
            case CacheDriver.Redis:
                return "infrahub.services.adapters.cache.redis"

    @property
    def driver_class_name(self) -> str:
        match self:
            case CacheDriver.NATS:
                return "NATSCache"
            case CacheDriver.Redis:
                return "RedisCache"


class WorkflowDriver(str, Enum):
    LOCAL = "local"
    WORKER = "worker"


class ExtraLogLevel(str, Enum):
    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"


class MainSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_")
    docs_index_path: Path = Field(
        default=Path("/opt/infrahub/docs/build/search-index.json"),
        description="Full path of saved json containing pre-indexed documentation",
    )
    internal_address: str | None = Field(default=None)
    allow_anonymous_access: bool = Field(
        default=True, description="Indicates if the system allows anonymous read access"
    )
    anonymous_access_role: str = Field(
        default="Anonymous User", description="Name of the role defining which permissions anonymous users have"
    )
    telemetry_optout: bool = Field(default=False, description="Disable anonymous usage reporting")
    telemetry_endpoint: str = "https://telemetry.opsmill.cloud/infrahub"
    permission_backends: list[str] = Field(
        default=["infrahub.permissions.LocalPermissionBackend"],
        description="List of modules to handle permissions, they will be run in the given order",
    )
    public_url: str | None = Field(
        default=None,
        description="Define the public URL of the Infrahub, might be required for OAuth2 and OIDC depending on your infrastructure.",
    )
    schema_strict_mode: bool = Field(
        default=True,
        description="Enable strict schema validation. When set to `False`, "
        "`human_friendly_id` schema fields should not necessarily target a unique combination of peer attributes.",
    )

    @field_validator("docs_index_path", mode="before")
    @classmethod
    def convert_to_path(cls, value: Path | str) -> Path:
        return Path(value) if isinstance(value, str) else value


class FileSystemStorageSettings(BaseSettings):
    # Make variable lookup case-sensitive to avoid fetching $PATH value
    model_config = SettingsConfigDict(case_sensitive=True)
    path_: Path = Field(
        default=Path("/opt/infrahub/storage"),
        alias="path",
        validation_alias=AliasChoices("INFRAHUB_STORAGE_LOCAL_PATH", "infrahub_storage_local_path", "path"),
    )

    @field_validator("path_", mode="before")
    @classmethod
    def convert_to_path(cls, value: Path | str) -> Path:
        return Path(value) if isinstance(value, str) else value


class S3StorageSettings(BaseSettings):
    access_key_id: str = Field(default="", alias="AWS_ACCESS_KEY_ID", validation_alias="AWS_ACCESS_KEY_ID")
    secret_access_key: str = Field(default="", alias="AWS_SECRET_ACCESS_KEY", validation_alias="AWS_SECRET_ACCESS_KEY")
    bucket_name: str = Field(
        default="",
        alias="AWS_S3_BUCKET_NAME",
        validation_alias=AliasChoices("INFRAHUB_STORAGE_BUCKET_NAME", "AWS_S3_BUCKET_NAME"),
    )
    endpoint_url: str = Field(
        default="",
        alias="AWS_S3_ENDPOINT_URL",
        validation_alias=AliasChoices("INFRAHUB_STORAGE_ENDPOINT_URL", "AWS_S3_ENDPOINT_URL"),
    )
    use_ssl: bool = Field(
        default=True,
        alias="AWS_S3_USE_SSL",
        validation_alias=AliasChoices("INFRAHUB_STORAGE_USE_SSL", "AWS_S3_USE_SSL"),
    )
    default_acl: str = Field(
        default="private",
        alias="AWS_DEFAULT_ACL",
        validation_alias=AliasChoices("INFRAHUB_STORAGE_DEFAULT_ACL", "AWS_DEFAULT_ACL"),
    )
    querystring_auth: bool = Field(
        default=False,
        alias="AWS_QUERYSTRING_AUTH",
        validation_alias=AliasChoices("INFRAHUB_STORAGE_QUERYSTRING_AUTH", "AWS_QUERYSTRING_AUTH"),
    )
    custom_domain: str = Field(
        default="",
        alias="AWS_S3_CUSTOM_DOMAIN",
        validation_alias=AliasChoices("INFRAHUB_STORAGE_CUSTOM_DOMAIN", "AWS_S3_CUSTOM_DOMAIN"),
    )


class StorageSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_STORAGE_")
    driver: StorageDriver = StorageDriver.FileSystemStorage
    local: FileSystemStorageSettings = FileSystemStorageSettings()
    s3: S3StorageSettings = S3StorageSettings()


class DatabaseSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_DB_")
    db_type: DatabaseType = Field(
        default=DatabaseType.NEO4J, validation_alias=AliasChoices("INFRAHUB_DB_TYPE", "db_type")
    )
    protocol: str = "bolt"
    username: str = "neo4j"
    password: str = "admin"
    address: str = "localhost"
    port: int = 7687
    database: str | None = Field(default=None, pattern=VALID_DATABASE_NAME_REGEX, description="Name of the database")
    tls_enabled: bool = Field(default=False, description="Indicates if TLS is enabled for the connection")
    tls_insecure: bool = Field(default=False, description="Indicates if TLS certificates are verified")
    tls_ca_file: str | None = Field(default=None, description="File path to CA cert or bundle in PEM format")
    query_size_limit: int = Field(
        default=5_000,
        ge=1,
        le=20_000,
        description="The max number of records to fetch in a single query before performing internal pagination.",
    )
    max_depth_search_hierarchy: int = Field(
        default=5,
        le=20,
        description="Maximum number of level to search in a hierarchy.",
    )
    retry_limit: int = Field(
        default=3, description="Maximum number of times a transient issue in a transaction should be retried."
    )
    max_concurrent_queries: int = Field(
        default=0, ge=0, description="Maximum number of concurrent queries that can run (0 means unlimited)."
    )
    max_concurrent_queries_delay: float = Field(
        default=0.01, ge=0, description="Delay to add when max_concurrent_queries is reached."
    )

    @property
    def database_name(self) -> str:
        return self.database or self.db_type.value


class DevelopmentSettings(BaseSettings):
    """The development settings are only relevant for local development"""

    model_config = SettingsConfigDict(env_prefix="INFRAHUB_DEV_")

    frontend_redirect_sso: bool = Field(
        default=False,
        description="Indicates of the frontend should be responsible for the SSO redirection",
    )


class BrokerSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_BROKER_")
    enable: bool = True
    tls_enabled: bool = Field(default=False, description="Indicates if TLS is enabled for the connection")
    tls_insecure: bool = Field(default=False, description="Indicates if TLS certificates are verified")
    tls_ca_file: str | None = Field(default=None, description="File path to CA cert or bundle in PEM format")
    username: str = "infrahub"
    password: str = "infrahub"
    address: str = "localhost"
    port: int | None = Field(default=None, ge=1, le=65535, description="Specified if running on a non default port.")
    rabbitmq_http_port: int | None = Field(default=None, ge=1, le=65535)
    namespace: str = "infrahub"
    maximum_message_retries: int = Field(
        default=10, description="The maximum number of retries that are attempted for failed messages"
    )
    maximum_concurrent_messages: int = Field(
        default=2, description="The maximum number of concurrent messages fetched by each worker", ge=1
    )
    virtualhost: str = Field(default="/", description="The virtual host to connect to")
    driver: BrokerDriver = BrokerDriver.RabbitMQ

    @property
    def service_port(self) -> int:
        default_ports: dict[bool, int] = {True: 5671, False: 5672}
        if self.driver == BrokerDriver.NATS:
            return self.port or 4222
        return self.port or default_ports[self.tls_enabled]


class CacheSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_CACHE_")
    enable: bool = True
    address: str = "localhost"
    port: int | None = Field(
        default=None, ge=1, le=65535, description="Specified if running on a non default port (6379)"
    )
    database: int = Field(default=0, ge=0, le=15, description="Id of the database to use")
    driver: CacheDriver = CacheDriver.Redis
    username: str = ""
    password: str = ""
    tls_enabled: bool = Field(default=False, description="Indicates if TLS is enabled for the connection")
    tls_insecure: bool = Field(default=False, description="Indicates if TLS certificates are verified")
    tls_ca_file: str | None = Field(default=None, description="File path to CA cert or bundle in PEM format")

    @property
    def service_port(self) -> int:
        default_ports: int = 6379
        if self.driver == CacheDriver.NATS:
            return self.port or 4222
        return self.port or default_ports


class WorkflowSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_WORKFLOW_")
    enable: bool = True
    address: str = "localhost"
    port: int | None = Field(default=None, ge=1, le=65535, description="Specified if running on a non default port.")
    tls_enabled: bool = Field(default=False, description="Indicates if TLS is enabled for the connection")
    driver: WorkflowDriver = WorkflowDriver.WORKER
    default_worker_type: str = "infrahubasync"
    extra_loggers: list[str] = Field(
        default_factory=list, description="A list of additional logger that will be captured during task execution."
    )
    extra_log_level: ExtraLogLevel = Field(
        default=ExtraLogLevel.INFO, description="Log level applied to all extra loggers."
    )
    worker_polling_interval: int = Field(
        default=2, ge=1, le=30, description="Specify how often the worker should poll the server for tasks (sec)"
    )

    @property
    def api_endpoint(self) -> str:
        url = "https://" if self.tls_enabled else "http://"
        url += self.address
        if self.port:
            url += f":{self.port}"
        url += "/api"
        return url


class ApiSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_API_")
    cors_allow_origins: list[str] = Field(
        default_factory=list, description="A list of origins that are authorized to make cross-site HTTP requests"
    )
    cors_allow_methods: list[str] = Field(
        default_factory=default_cors_allow_methods,
        description="A list of HTTP verbs that are allowed for the actual request",
    )
    cors_allow_headers: list[str] = Field(
        default_factory=default_cors_allow_headers,
        description="The list of non-standard HTTP headers allowed in requests from the browser",
    )
    cors_allow_credentials: bool = Field(
        default=True, description="If True, cookies will be allowed to be included in cross-site HTTP requests"
    )


class GitSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_GIT_")
    repositories_directory: str = "repositories"
    sync_interval: int = Field(
        default=10,
        ge=0,
        description="Time (in seconds) between git repositories synchronizations",
        deprecated="This setting is deprecated and not currently in use.",
    )
    append_git_suffix: list[str] = Field(
        default_factory=default_append_git_suffix_domains,
        description="Automatically append '.git' to HTTP URLs if for these domains.",
    )


class HTTPSettings(BaseSettings):
    """The HTTP settings control how Infrahub interacts with external HTTP servers

    This can be things like webhooks and OAuth2 providers"""

    model_config = SettingsConfigDict(env_prefix="INFRAHUB_HTTP_")
    timeout: int = Field(default=10, description="Default connection timeout in seconds")
    tls_insecure: bool = Field(
        default=False,
        description="Indicates if Infrahub will validate server certificates or if the validation is ignored.",
    )
    tls_ca_bundle: str | None = Field(
        default=None,
        description="Custom CA bundle in PEM format. The value should either be the CA bundle as a string, alternatively as a file path.",
    )

    @model_validator(mode="after")
    def set_tls_context(self) -> Self:
        try:
            # Validate that the context can be created, we want to raise this error during application start
            # instead of running into issues later when we first try to use the tls context.
            self.get_tls_context()
        except ssl.SSLError as exc:
            raise ValueError(f"Unable load CA bundle from {self.tls_ca_bundle}: {exc}") from exc

        return self

    def get_tls_context(self) -> ssl.SSLContext:
        if self.tls_insecure:
            return ssl._create_unverified_context()

        if not self.tls_ca_bundle:
            return ssl.create_default_context()

        tls_ca_path = Path(self.tls_ca_bundle)

        try:
            possibly_file = tls_ca_path.exists()
        except OSError:
            # Raised if the filename is too long which can indicate
            # that the value is a PEM certificate in string form.
            possibly_file = False

        if possibly_file and tls_ca_path.is_file():
            context = ssl.create_default_context(cafile=str(tls_ca_path))
        else:
            context = ssl.create_default_context()
            context.load_verify_locations(cadata=self.tls_ca_bundle)

        return context


class InitialSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_INITIAL_")
    default_branch: str = Field(
        default="main",
        description="Defines the name of the default branch within Infrahub, can only be set once during initialization of the system.",
    )
    admin_token: str | None = Field(default=None, description="An optional initial token for the admin account.")
    admin_password: str = Field(default="infrahub", description="The initial password for the admin user")
    agent_token: str | None = Field(default=None, description="An optional initial token for a git-agent account.")
    agent_password: str | None = Field(
        default=None, description="An optional initial password for a git-agent account."
    )

    @property
    def create_agent_user(self) -> bool:
        if self.agent_token or self.agent_password:
            return True
        return False

    @model_validator(mode="after")
    def check_tokens_match(self) -> Self:
        if self.admin_token is not None and self.agent_token is not None and self.admin_token == self.agent_token:
            raise ValueError("Initial user tokens can't have the same values")
        return self


def _default_scopes() -> list[str]:
    return ["openid", "profile", "email"]


class SecurityOIDCBaseSettings(BaseSettings):
    """Baseclass for typing"""

    icon: str = Field(default="mdi:account-key")
    display_label: str = Field(default="Single Sign on")
    userinfo_method: UserInfoMethod = Field(default=UserInfoMethod.GET)


class SecurityOIDCSettings(SecurityOIDCBaseSettings):
    client_id: str = Field(..., description="Client ID of the application created in the auth provider")
    client_secret: str = Field(..., description="Client secret as defined in auth provider")
    discovery_url: str = Field(..., description="The OIDC discovery URL xyz/.well-known/openid-configuration")
    scopes: list[str] = Field(default_factory=_default_scopes)


class SecurityOIDCGoogle(SecurityOIDCSettings):
    """Settings for the custom OIDC provider"""

    model_config = SettingsConfigDict(env_prefix="INFRAHUB_OIDC_GOOGLE_")

    discovery_url: str = Field(default="https://accounts.google.com/.well-known/openid-configuration")
    icon: str = Field(default="mdi:google")
    display_label: str = Field(default="Google")


class SecurityOIDCProvider1(SecurityOIDCSettings):
    """Settings for the custom OIDC provider"""

    model_config = SettingsConfigDict(env_prefix="INFRAHUB_OIDC_PROVIDER1_")


class SecurityOIDCProvider2(SecurityOIDCSettings):
    """Settings for the custom OIDC provider"""

    model_config = SettingsConfigDict(env_prefix="INFRAHUB_OIDC_PROVIDER2_")


class SecurityOIDCProviderSettings(BaseModel):
    """This class is meant to facilitate configuration of OIDC providers when loading configuration from a infrahub.toml file."""

    google: SecurityOIDCGoogle | None = Field(default=None)
    provider1: SecurityOIDCProvider1 | None = Field(default=None)
    provider2: SecurityOIDCProvider2 | None = Field(default=None)


class SecurityOAuth2BaseSettings(BaseSettings):
    """Baseclass for typing"""

    icon: str = Field(default="mdi:account-key")
    userinfo_method: UserInfoMethod = Field(default=UserInfoMethod.GET)


class SecurityOAuth2Settings(SecurityOAuth2BaseSettings):
    """Common base for Oauth2 providers"""

    client_id: str = Field(..., description="Client ID of the application created in the auth provider")
    client_secret: str = Field(..., description="Client secret as defined in auth provider")
    authorization_url: str = Field(...)
    token_url: str = Field(...)
    userinfo_url: str = Field(...)
    scopes: list[str] = Field(default_factory=_default_scopes)
    display_label: str = Field(default="Single Sign on")


class SecurityOAuth2Provider1(SecurityOAuth2Settings):
    """Common base for Oauth2 providers"""

    model_config = SettingsConfigDict(env_prefix="INFRAHUB_OAUTH2_PROVIDER1_")


class SecurityOAuth2Provider2(SecurityOAuth2Settings):
    """Common base for Oauth2 providers"""

    model_config = SettingsConfigDict(env_prefix="INFRAHUB_OAUTH2_PROVIDER2_")


class SecurityOAuth2Google(SecurityOAuth2Settings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_OAUTH2_GOOGLE_")
    authorization_url: str = Field(default="https://accounts.google.com/o/oauth2/auth")
    token_url: str = Field(default="https://oauth2.googleapis.com/token")
    userinfo_url: str = Field(default="https://www.googleapis.com/oauth2/v3/userinfo")
    icon: str = Field(default="mdi:google")
    display_label: str = Field(default="Google")


class SecurityOAuth2ProviderSettings(BaseModel):
    """This class is meant to facilitate configuration of OAuth2 providers when loading configuration from a infrahub.toml file."""

    google: SecurityOAuth2Google | None = Field(default=None)
    provider1: SecurityOAuth2Provider1 | None = Field(default=None)
    provider2: SecurityOAuth2Provider2 | None = Field(default=None)


class MiscellaneousSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_MISC_")
    print_query_details: bool = False
    start_background_runner: bool = True
    maximum_validator_execution_time: int = Field(
        default=1800, description="The maximum allowed time (in seconds) for a validator to run."
    )
    response_delay: int = Field(default=0, description="Arbitrary delay to add when processing API requests.")


class RemoteLoggingSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_LOGGING_REMOTE_")
    enable: bool = False
    frontend_dsn: str | None = None
    api_server_dsn: str | None = None
    git_agent_dsn: str | None = None


class LoggingSettings(BaseSettings):
    remote: RemoteLoggingSettings = RemoteLoggingSettings()


class AnalyticsSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_ANALYTICS_")
    enable: bool = True
    address: str | None = None
    api_key: str | None = None


class ExperimentalFeaturesSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_EXPERIMENTAL_")
    graphql_enums: bool = False
    value_db_index: bool = False


class SecuritySettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_SECURITY_")
    access_token_lifetime: int = Field(default=3600, description="Lifetime of access token in seconds")
    refresh_token_lifetime: int = Field(
        default=THIRTY_DAYS_IN_SECONDS, description="Lifetime of refresh token in seconds"
    )
    secret_key: str = Field(
        default_factory=generate_uuid, description="The secret key used to validate authentication tokens"
    )
    oauth2_providers: list[Oauth2Provider] = Field(default_factory=list, description="The selected OAuth2 providers")
    oauth2_provider_settings: SecurityOAuth2ProviderSettings = Field(default_factory=SecurityOAuth2ProviderSettings)
    oidc_providers: list[OIDCProvider] = Field(default_factory=list, description="The selected OIDC providers")
    oidc_provider_settings: SecurityOIDCProviderSettings = Field(default_factory=SecurityOIDCProviderSettings)
    restrict_untrusted_jinja2_filters: bool = Field(
        default=True, description="Indicates if untrusted Jinja2 filters should be disallowd for computed attributes"
    )
    _oauth2_settings: dict[str, SecurityOAuth2Settings] = PrivateAttr(default_factory=dict)
    _oidc_settings: dict[str, SecurityOIDCSettings] = PrivateAttr(default_factory=dict)
    sso_user_default_group: str | None = Field(
        default=None,
        description="Name of the group to which users authenticated via SSO will belong if not provided by identity provider",
    )

    @model_validator(mode="after")
    def check_oauth2_provider_settings(self) -> Self:
        mapped_providers: dict[Oauth2Provider, type[SecurityOAuth2BaseSettings]] = {
            Oauth2Provider.PROVIDER1: SecurityOAuth2Provider1,
            Oauth2Provider.PROVIDER2: SecurityOAuth2Provider2,
            Oauth2Provider.GOOGLE: SecurityOAuth2Google,
        }
        for oauth2_provider in self.oauth2_providers:
            match oauth2_provider:
                case Oauth2Provider.GOOGLE:
                    if self.oauth2_provider_settings.google:
                        self._oauth2_settings[oauth2_provider.value] = self.oauth2_provider_settings.google
                case Oauth2Provider.PROVIDER1:
                    if self.oauth2_provider_settings.provider1:
                        self._oauth2_settings[oauth2_provider.value] = self.oauth2_provider_settings.provider1
                case Oauth2Provider.PROVIDER2:
                    if self.oauth2_provider_settings.provider2:
                        self._oauth2_settings[oauth2_provider.value] = self.oauth2_provider_settings.provider2

            if oauth2_provider.value not in self._oauth2_settings:
                provider = mapped_providers[oauth2_provider]()
                if isinstance(provider, SecurityOAuth2Settings):
                    self._oauth2_settings[oauth2_provider.value] = provider

        return self

    @model_validator(mode="after")
    def check_oidc_provider_settings(self) -> Self:
        mapped_providers: dict[OIDCProvider, type[SecurityOIDCBaseSettings]] = {
            OIDCProvider.GOOGLE: SecurityOIDCGoogle,
            OIDCProvider.PROVIDER1: SecurityOIDCProvider1,
            OIDCProvider.PROVIDER2: SecurityOIDCProvider2,
        }
        for oidc_provider in self.oidc_providers:
            match oidc_provider:
                case OIDCProvider.GOOGLE:
                    if self.oidc_provider_settings.google:
                        self._oidc_settings[oidc_provider.value] = self.oidc_provider_settings.google
                case OIDCProvider.PROVIDER1:
                    if self.oidc_provider_settings.provider1:
                        self._oidc_settings[oidc_provider.value] = self.oidc_provider_settings.provider1
                case OIDCProvider.PROVIDER2:
                    if self.oidc_provider_settings.provider2:
                        self._oidc_settings[oidc_provider.value] = self.oidc_provider_settings.provider2

            if oidc_provider.value not in self._oidc_settings:
                provider = mapped_providers[oidc_provider]()
                if isinstance(provider, SecurityOIDCSettings):
                    self._oidc_settings[oidc_provider.value] = provider

        return self

    def get_oauth2_provider(self, provider: str) -> SecurityOAuth2Settings:
        if provider in self._oauth2_settings:
            return self._oauth2_settings[provider]

        raise ProcessingError(message=f"The provider {provider} has not been initialized")

    def get_oidc_provider(self, provider: str) -> SecurityOIDCSettings:
        if provider in self._oidc_settings:
            return self._oidc_settings[provider]

        raise ProcessingError(message=f"The provider {provider} has not been initialized")

    @property
    def public_sso_config(self) -> SSOInfo:
        oauth2_providers = [
            SSOProviderInfo(
                name=provider,
                display_label=self._oauth2_settings[provider].display_label,
                icon=self._oauth2_settings[provider].icon,
                protocol=SSOProtocol.OAUTH2,
            )
            for provider in self._oauth2_settings
        ]
        oidc_providers = [
            SSOProviderInfo(
                name=provider,
                display_label=self._oidc_settings[provider].display_label,
                icon=self._oidc_settings[provider].icon,
                protocol=SSOProtocol.OIDC,
            )
            for provider in self._oidc_settings
        ]
        return SSOInfo(providers=oauth2_providers + oidc_providers)


class TraceSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_TRACE_")
    enable: bool = Field(default=False)
    insecure: bool = Field(
        default=True, description="Use insecure connection (HTTP) if True, otherwise use secure connection (HTTPS)"
    )
    exporter_type: TraceExporterType = Field(
        default=TraceExporterType.CONSOLE, description="Type of exporter to be used for tracing"
    )
    exporter_protocol: TraceTransportProtocol = Field(
        default=TraceTransportProtocol.GRPC, description="Protocol to be used for exporting traces"
    )
    exporter_endpoint: str | None = Field(default=None, description="OTLP endpoint for exporting traces")


@dataclass
class Override:
    message_bus: InfrahubMessageBus | None = None
    cache: InfrahubCache | None = None
    workflow: InfrahubWorkflow | None = None


@dataclass
class ConfiguredSettings:
    settings: Settings | None = None

    def initialize(self, config_file: Path | str | None = None) -> None:
        """Initialize the settings if they have not been initialized."""
        if self.initialized:
            return
        if not config_file:
            config_file_name: str = os.environ.get("INFRAHUB_CONFIG", "infrahub.toml")
            config_file = Path(config_file_name).resolve()
        self.settings = load(config_file)

    def initialize_and_exit(self, config_file: Path | str | None = None) -> None:
        """Initialize the settings if they have not been initialized, exit on failures."""
        if self.initialized:
            return
        if not config_file:
            config_file_name = os.environ.get("INFRAHUB_CONFIG", "infrahub.toml")
            config_file = Path(config_file_name).resolve()
        load_and_exit(config_file)

    @property
    def active_settings(self) -> Settings:
        if self.settings:
            return self.settings
        raise InitializationError

    @property
    def initialized(self) -> bool:
        return self.settings is not None

    @property
    def main(self) -> MainSettings:
        return self.active_settings.main

    @property
    def api(self) -> ApiSettings:
        return self.active_settings.api

    @property
    def git(self) -> GitSettings:
        return self.active_settings.git

    @property
    def http(self) -> HTTPSettings:
        return self.active_settings.http

    @property
    def database(self) -> DatabaseSettings:
        return self.active_settings.database

    @property
    def broker(self) -> BrokerSettings:
        return self.active_settings.broker

    @property
    def cache(self) -> CacheSettings:
        return self.active_settings.cache

    @property
    def dev(self) -> DevelopmentSettings:
        return self.active_settings.dev

    @property
    def workflow(self) -> WorkflowSettings:
        return self.active_settings.workflow

    @property
    def miscellaneous(self) -> MiscellaneousSettings:
        return self.active_settings.miscellaneous

    @property
    def initial(self) -> InitialSettings:
        return self.active_settings.initial

    @property
    def logging(self) -> LoggingSettings:
        return self.active_settings.logging

    @property
    def analytics(self) -> AnalyticsSettings:
        return self.active_settings.analytics

    @property
    def security(self) -> SecuritySettings:
        return self.active_settings.security

    @property
    def storage(self) -> StorageSettings:
        return self.active_settings.storage

    @property
    def trace(self) -> TraceSettings:
        return self.active_settings.trace

    @property
    def experimental_features(self) -> ExperimentalFeaturesSettings:
        return self.active_settings.experimental_features


class Settings(BaseSettings):
    """Main Settings Class for the project."""

    main: MainSettings = MainSettings()
    api: ApiSettings = ApiSettings()
    git: GitSettings = GitSettings()
    dev: DevelopmentSettings = DevelopmentSettings()
    http: HTTPSettings = HTTPSettings()
    database: DatabaseSettings = DatabaseSettings()
    broker: BrokerSettings = BrokerSettings()
    cache: CacheSettings = CacheSettings()
    workflow: WorkflowSettings = WorkflowSettings()
    miscellaneous: MiscellaneousSettings = MiscellaneousSettings()
    logging: LoggingSettings = LoggingSettings()
    analytics: AnalyticsSettings = AnalyticsSettings()
    initial: InitialSettings = InitialSettings()
    security: SecuritySettings = SecuritySettings()
    storage: StorageSettings = StorageSettings()
    trace: TraceSettings = TraceSettings()
    experimental_features: ExperimentalFeaturesSettings = ExperimentalFeaturesSettings()


def load(config_file_name: Path | str = "infrahub.toml", config_data: dict[str, Any] | None = None) -> Settings:
    """Load configuration.

    Configuration is loaded from a config file in toml format that contains the settings,
    or from a dictionary of those settings passed in as "config_data"
    """
    config_file = Path(config_file_name)

    if config_data:
        return Settings(**config_data)

    if config_file.exists():
        config_string = config_file.read_text(encoding="utf-8")
        config_tmp = toml.loads(config_string)

        return Settings(**config_tmp)

    return Settings()


def load_and_exit(config_file_name: Path | str = "infrahub.toml", config_data: dict[str, Any] | None = None) -> None:
    """Calls load, but wraps it in a try except block.

    This is done to handle a ValidationError which is raised when settings are specified but invalid.
    In such cases, a message is printed to the screen indicating the settings which don't pass validation.

    Args:
        config_file_name (str, optional): [description]. Defaults to "pyproject.toml".
        config_data (dict, optional): [description]. Defaults to None.
    """
    try:
        SETTINGS.settings = load(config_file_name=config_file_name, config_data=config_data)
    except ValidationError as err:
        print(f"Configuration not valid, found {len(err.errors())} error(s)")
        for error in err.errors():
            error_locations = [str(location) for location in error["loc"]]
            print(f"  {'/'.join(error_locations)} | {error['msg']} ({error['type']})")
        sys.exit(1)


OVERRIDE: Override = Override()
SETTINGS: ConfiguredSettings = ConfiguredSettings()



from typing import TYPE_CHECKING, Sequence, cast

from prefect.server.database import PrefectDBInterface, db_injector
from prefect.server.events.filters import EventFilter, EventNameFilter, EventOrder, EventRelatedFilter
from prefect.server.events.schemas.events import ReceivedEvent
from prefect.server.utilities.schemas import PrefectBaseModel
from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from sqlalchemy.sql.expression import ColumnExpressionArgument


class InfrahubEventFilter(EventFilter):
    matching_related: list[EventRelatedFilter] = Field(default_factory=list)

    def set_prefix(self) -> None:
        if self.event:
            if self.event.prefix is not None and "infrahub." not in self.event.prefix:
                self.event.prefix.append("infrahub.")
        else:
            self.event = EventNameFilter(prefix=["infrahub."], name=[], exclude_prefix=None, exclude_name=None)

    @db_injector
    def build_where_clauses(self, db: PrefectDBInterface) -> Sequence["ColumnExpressionArgument[bool]"]:
        result = cast(list["ColumnExpressionArgument[bool]"], super().build_where_clauses())
        top_level_filter = self._scoped_event_resources(db)
        for matching_related in self.matching_related:
            matching_related._top_level_filter = top_level_filter
            result.extend(matching_related.build_where_clauses())

        return result

    @classmethod
    def default(cls) -> "InfrahubEventFilter":
        return cls(event=None, any_resource=None, resource=None, related=None, order=EventOrder.DESC)


class InfrahubEventPage(PrefectBaseModel):
    events: list[ReceivedEvent] = Field(..., description="The Events matching the query")
    total: int = Field(..., description="The total number of matching Events")


class InfrahubEventfilterInput(BaseModel):
    limit: int = Field(default=50)
    filter: InfrahubEventFilter = Field(default_factory=InfrahubEventFilter.default)
    offset: int | None = Field(default=None)



from __future__ import annotations

from fastapi import APIRouter, FastAPI
from prefect.server.api.server import create_app

from . import events

router = APIRouter(prefix="/infrahub")

router.include_router(events.router)


def create_infrahub_prefect() -> FastAPI:
    app = create_app()
    api_app: FastAPI = app.__dict__["api_app"]
    api_app.include_router(router=router)

    return app



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import GraphMigration, MigrationResult
from infrahub.log import get_logger

from ...constants import GLOBAL_BRANCH_NAME, BranchSupportType
from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class FixBranchAwareEdgesQuery(Query):
    name = "replace_global_edges"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        """
        Between a Node and a Relationship, if Relationship.branch_support=aware, replace any global edge
        to the branch of a non-global edge leaving out of the Relationship node. Note that there can't
        be multiple non-global branches on these edges, as a dedicated Relationship node would exist for that.
        """

        query = """
        MATCH (node:Node)-[global_edge:IS_RELATED {branch: $global_branch}]-(rel:Relationship)
        WHERE rel.branch_support=$branch_aware
        MATCH (rel)-[non_global_edge:IS_RELATED]-(node_2: Node)
        WHERE non_global_edge.branch <> $global_branch
        SET global_edge.branch = non_global_edge.branch
        """

        params = {
            "global_branch": GLOBAL_BRANCH_NAME,
            "branch_aware": BranchSupportType.AWARE.value,
            "branch_agnostic": BranchSupportType.AGNOSTIC.value,
        }

        self.params.update(params)
        self.add_to_query(query)


class SetMissingToTimeQuery(Query):
    name = "set_missing_to_time"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        """
        If both a deleted edge and an active edge with no time exist between 2 nodes on the same branch,
        set `to` time of active edge using `from` time of the deleted one. This would typically happen after having
        replaced a deleted edge on global branch by correct branch with above query.
        """

        query = """
        MATCH (node:Node)-[deleted_edge:IS_RELATED {status: "deleted"}]-(rel:Relationship)
        MATCH (rel)-[active_edge:IS_RELATED {status: "active"}]-(node)
        WHERE active_edge.to IS NULL AND deleted_edge.branch = active_edge.branch
        SET active_edge.to = deleted_edge.from
        """

        self.add_to_query(query)


class DeleteNodesRelsQuery(Query):
    name = "delete_relationships_of_deleted_nodes"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        """
        Some nodes may have been incorrectly deleted, typically, while these nodes edges connected to Root
        are correctly deleted, edges connected to other `Node` through a `Relationship` node may still be active.
        Following query correctly deletes these edges by both setting correct to time and creating corresponding deleted edge.
        """

        query = """
        MATCH (deleted_node: Node)-[deleted_edge:IS_PART_OF {status: "deleted"}]->(:Root)
        MATCH (deleted_node)-[:IS_RELATED]-(rel:Relationship)

        // exclude nodes having been deleted through migration. find those with same uuid and exclude the one with earlier
        // timestamp on active branch
        WHERE NOT EXISTS {
          MATCH (deleted_node)-[e1:IS_RELATED]-(rel)-[e2:IS_RELATED]-(other_node)
          WITH deleted_node, other_node, MIN(e1.from) AS min_e1_from, MIN(e2.from) AS min_e2_from
          WHERE deleted_node <> other_node AND deleted_node.uuid = other_node.uuid AND min_e1_from < min_e2_from
        }

        // Note that if an AWARE node has been deleted on a branch and relationship is AGNOSTIC, we do not "delete" this relationship
        // right now as this aware node might exist on another branch.

        // Set to time if there is an active edge:
        // - on deleted edge branch
        // - or on any branch and deleted node is agnostic
        // - or deleted node is aware and rel is agnostic
        CALL {
            WITH rel, deleted_edge
            OPTIONAL MATCH (rel)-[peer_active_edge {status: "active"}]-(peer_1)
            WHERE (peer_active_edge.branch = deleted_edge.branch OR (rel.branch_support <> $branch_agnostic AND deleted_edge.branch = $global_branch))
            AND peer_active_edge.to IS NULL
            SET peer_active_edge.to = deleted_edge.from
        }

        // Get distinct rel nodes linked to a deleted node, with the time at which we should delete rel edges.
        // Take the MAX time so if it does not take the deleted time of a node deleted through a duplication migration.
        WITH DISTINCT rel,
            deleted_edge.branch AS deleted_edge_branch,
            deleted_edge.branch_level AS branch_level,
            MAX(deleted_edge.from) as deleted_time,
            deleted_node.branch_support as deleted_node_branch_support


        // No need to check deleted edge branch because
        // If deleted_node has different branch support type (agnostic/aware) than rel type,
        // there might already be a deleted edge that we would not match if we filter on deleted_edge_branch.
        // If both are aware, it still works, as we would have one Relationship node for each branch on which this relationship exists.
        MATCH (rel)-[]-(peer_2)
        WHERE NOT exists((rel)-[{status: "deleted"}]-(peer_2))


        // If res is agnostic and delete node is agnostic, we should delete on global branch
        // If rel is aware and deleted node is aware, we should use deleted edge branch
        // If rel is aware and delete node is agnostic, we need to create deleted edges for every distinct branch on which this relationship exists.
        WITH DISTINCT
            CASE
                // Branch on which `deleted` edge should be created depends on rel.branch_support.
                WHEN rel.branch_support = $branch_agnostic
                THEN CASE
                    WHEN deleted_node_branch_support = $branch_agnostic THEN [$global_branch]
                    ELSE []
                END
                ELSE
                CASE
                    WHEN deleted_node_branch_support = $branch_agnostic
                    THEN COLLECT {
                        WITH rel
                        MATCH (rel)-[active_edge {status: "active"}]-(peer_2)
                        RETURN DISTINCT active_edge.branch
                    }
                    ELSE
                    CASE
                        // if no active edge on this branch exists it means this relationship node is dedicated for another branch
                        WHEN exists((rel)-[{status: "active", branch: deleted_edge_branch}]-(peer_2)) THEN [deleted_edge_branch]
                        ELSE []
                    END
                END
            END AS branches,
          branch_level,
          deleted_time,
          peer_2,
          rel

        UNWIND branches as branch

        // Then creates `deleted` edge.
        // Below CALL subqueries are called once for each rel-peer_2 pair for which we want to create a deleted edge.
        // Note that with current infrahub relationships edges design, only one of this CALL should be matched per pair.

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)-[:IS_RELATED]->(peer_2)
            MERGE (rel)-[:IS_RELATED {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]->(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)-[:IS_PROTECTED]->(peer_2)
            MERGE (rel)-[:IS_PROTECTED {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]->(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)-[:IS_VISIBLE]->(peer_2)
            MERGE (rel)-[:IS_VISIBLE {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]->(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)-[:HAS_OWNER]->(peer_2)
            MERGE (rel)-[:HAS_OWNER {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]->(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)-[:HAS_SOURCE]->(peer_2)
            MERGE (rel)-[:HAS_SOURCE {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]->(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)<-[:IS_RELATED]-(peer_2)
            MERGE (rel)<-[:IS_RELATED {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]-(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)<-[:IS_PROTECTED]-(peer_2)
            MERGE (rel)<-[:IS_PROTECTED {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]-(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)<-[:IS_VISIBLE]-(peer_2)
            MERGE (rel)<-[:IS_VISIBLE {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]-(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)<-[:HAS_OWNER]-(peer_2)
            MERGE (rel)<-[:HAS_OWNER {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]-(peer_2)
        }

        CALL {
            WITH rel, peer_2, branch, branch_level, deleted_time
            MATCH (rel)<-[:HAS_SOURCE]-(peer_2)
            MERGE (rel)<-[:HAS_SOURCE {status: "deleted", branch: branch, branch_level: branch_level, from: deleted_time}]-(peer_2)
        }
        """

        params = {
            "global_branch": GLOBAL_BRANCH_NAME,
            "branch_aware": BranchSupportType.AWARE.value,
            "branch_agnostic": BranchSupportType.AGNOSTIC.value,
        }

        self.params.update(params)
        self.add_to_query(query)


class Migration019(GraphMigration):
    """
    Fix corrupted state introduced by Migration012 when duplicating a CoreAccount (branch Aware)
    being part of a CoreStandardGroup (branch Agnostic). Database is corrupted at multiple points:
    - Old CoreAccount node <> group_member node `active` edge has no `to` time (possibly because of #5590).
    - Old CoreAccount node <> group_member node `deleted` edge is on `$global_branch` branch instead of `main`.
    - New CoreAccount node <> group_member node `active` edge is on `$global_branch` branch instead of `main`.

    Also, users having deleted corresponding CoreStandardGroup will also have the following data corruption,
    as deletion did not happen correctly due to above issues:
    - Both CoreAccount <> group_member and CoreStandardGroup <> group_member edges
      have not been deleted (ie status is `active` without `to` time and no additional `deleted` edge).

    This migration fixes all above issues to have consistent edges, and fixes IFC-1204.
    """

    name: str = "019_fix_edges_state"
    minimum_version: int = 18
    queries: Sequence[type[Query]] = [FixBranchAwareEdgesQuery, SetMissingToTimeQuery, DeleteNodesRelsQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



from __future__ import annotations

from prefect import task
from prefect.cache_policies import NONE

from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreWebhook
from infrahub.database import InfrahubDatabase  # noqa: TC001  needed for prefect flow

from .models import WebhookTriggerDefinition


@task(name="gather-trigger-webhook", task_run_name="Gather webhook triggers", cache_policy=NONE)
async def gather_trigger_webhook(db: InfrahubDatabase) -> list[WebhookTriggerDefinition]:
    webhooks = await NodeManager.query(db=db, schema=CoreWebhook)
    triggers = [WebhookTriggerDefinition.from_object(webhook) for webhook in webhooks]
    return triggers



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

from infrahub.core.schema import AttributeSchema  # noqa: TC001

if TYPE_CHECKING:
    from infrahub.core.schema import GenericSchema, NodeSchema, SchemaAttributePath


@dataclass
class PythonDefinition:
    kind: str
    attribute: AttributeSchema

    @property
    def key_name(self) -> str:
        return f"{self.kind}_{self.attribute.name}"


class ComputedAttributeTarget(BaseModel):
    kind: str
    attribute: AttributeSchema
    filter_keys: list[str] = Field(default_factory=list)

    @property
    def key_name(self) -> str:
        return f"{self.kind}_{self.attribute.name}"

    @property
    def node_filters(self) -> list[str]:
        if self.filter_keys:
            return self.filter_keys

        return ["ids"]

    def __hash__(self) -> int:
        return hash((self.kind, self.attribute, tuple(self.filter_keys)))


class ComputedAttributeTriggerNode(BaseModel):
    kind: str
    attributes: list[str] = Field(default_factory=list)
    relationships: list[str] = Field(default_factory=list)
    targets_self: bool = Field(default=False)

    @property
    def fields(self) -> list[str]:
        return self.attributes + self.relationships


class RegisteredNodeComputedAttribute(BaseModel):
    local_fields: dict[str, list[ComputedAttributeTarget]] = Field(
        default_factory=dict,
        description="These are fields local to the modified node, which can include the names of attributes and relationships",
    )
    relationships: dict[str, list[ComputedAttributeTarget]] = Field(
        default_factory=dict,
        description="These relationships refer to the name of the relationship as seen from the source node.",
    )

    def get_targets(self, updates: list[str] | None = None) -> list[ComputedAttributeTarget]:
        targets: dict[str, ComputedAttributeTarget] = {}
        for attribute, entries in self.local_fields.items():
            if updates and attribute not in updates:
                continue

            for entry in entries:
                if entry.key_name not in targets:
                    targets[entry.key_name] = entry

        for relationship_name, entries in self.relationships.items():
            for entry in entries:
                filter_key = f"{relationship_name}__ids"
                if entry.key_name in targets and filter_key not in targets[entry.key_name].filter_keys:
                    targets[entry.key_name].filter_keys.append(filter_key)

        return list(targets.values())


class ComputedAttributes:
    def __init__(
        self,
        transform_attribute_map: dict[str, list[AttributeSchema]] | None = None,
        jinja2_attribute_map: dict[str, RegisteredNodeComputedAttribute] | None = None,
    ) -> None:
        self._computed_python_transform_attribute_map: dict[str, list[AttributeSchema]] = transform_attribute_map or {}
        self._computed_jinja2_attribute_map: dict[str, RegisteredNodeComputedAttribute] = jinja2_attribute_map or {}
        self._defined_from_generic: dict[str, str] = {}

    def duplicate(self) -> ComputedAttributes:
        return self.__class__(
            transform_attribute_map=deepcopy(self._computed_python_transform_attribute_map),
            jinja2_attribute_map=deepcopy(self._computed_jinja2_attribute_map),
        )

    def add_python_attribute(self, node: NodeSchema, attribute: AttributeSchema) -> None:
        if node.kind not in self._computed_python_transform_attribute_map:
            self._computed_python_transform_attribute_map[node.kind] = []
        self._computed_python_transform_attribute_map[node.kind].append(attribute)

    def get_kinds_python_attributes(self) -> list[str]:
        """Return kinds that have Python attributes defined"""
        return list(self._computed_python_transform_attribute_map.keys())

    def get_python_attributes_per_node(self) -> dict[str, list[AttributeSchema]]:
        return self._computed_python_transform_attribute_map

    @property
    def python_attributes_by_transform(self) -> dict[str, list[PythonDefinition]]:
        computed_attributes: dict[str, list[PythonDefinition]] = {}
        for kind, attributes in self._computed_python_transform_attribute_map.items():
            for attribute in attributes:
                if attribute.computed_attribute and attribute.computed_attribute.transform:
                    if attribute.computed_attribute.transform not in computed_attributes:
                        computed_attributes[attribute.computed_attribute.transform] = []

                    computed_attributes[attribute.computed_attribute.transform].append(
                        PythonDefinition(kind=kind, attribute=attribute)
                    )

        return computed_attributes

    def register_computed_jinja2(
        self, node: NodeSchema, attribute: AttributeSchema, schema_path: SchemaAttributePath
    ) -> None:
        key = node.kind
        if schema_path.is_type_relationship:
            key = schema_path.active_relationship_schema.peer

        if key not in self._computed_jinja2_attribute_map:
            self._computed_jinja2_attribute_map[key] = RegisteredNodeComputedAttribute()

        source_attribute = ComputedAttributeTarget(kind=node.kind, attribute=attribute)
        trigger_node = self._computed_jinja2_attribute_map[key]
        if schema_path.is_type_attribute:
            if schema_path.active_attribute_schema.name not in trigger_node.local_fields:
                trigger_node.local_fields[schema_path.active_attribute_schema.name] = []

            trigger_node.local_fields[schema_path.active_attribute_schema.name].append(deepcopy(source_attribute))
        elif schema_path.is_type_relationship:
            if schema_path.active_attribute_schema.name not in trigger_node.local_fields:
                trigger_node.local_fields[schema_path.active_attribute_schema.name] = []

            trigger_node.local_fields[schema_path.active_attribute_schema.name].append(deepcopy(source_attribute))

            if schema_path.active_relationship_schema.name not in trigger_node.relationships:
                trigger_node.relationships[schema_path.active_relationship_schema.name] = []

            trigger_node.relationships[schema_path.active_relationship_schema.name].append(deepcopy(source_attribute))

            if source_attribute.kind not in self._computed_jinja2_attribute_map:
                self._computed_jinja2_attribute_map[source_attribute.kind] = RegisteredNodeComputedAttribute()

            if (
                schema_path.active_relationship_schema.name
                not in self._computed_jinja2_attribute_map[source_attribute.kind].local_fields
            ):
                self._computed_jinja2_attribute_map[source_attribute.kind].local_fields[
                    schema_path.active_relationship_schema.name
                ] = []
            self._computed_jinja2_attribute_map[source_attribute.kind].local_fields[
                schema_path.active_relationship_schema.name
            ].append(deepcopy(source_attribute))

    def validate_generic_inheritance(
        self, node: NodeSchema, attribute: AttributeSchema, generic: GenericSchema
    ) -> None:
        attribute_key = f"{node.kind}__{attribute.name}"
        if duplicate := self._defined_from_generic.get(attribute_key):
            raise ValueError(
                f"{node.kind}: {attribute.name!r} is declared as a computed attribute from multiple generics {sorted([duplicate, generic.kind])}"
            )
        self._defined_from_generic[attribute_key] = generic.kind

    def get_impacted_jinja2_targets(self, kind: str, updates: list[str] | None = None) -> list[ComputedAttributeTarget]:
        if mapping := self._computed_jinja2_attribute_map.get(kind):
            return mapping.get_targets(updates=updates)

        return []

    def get_jinja2_target_map(self) -> dict[ComputedAttributeTarget, list[str]]:
        mapping: dict[ComputedAttributeTarget, set[str]] = {}

        for node, registered_computed_attribute in self._computed_jinja2_attribute_map.items():
            for local_fields in registered_computed_attribute.local_fields.values():
                for local_field in local_fields:
                    if local_field not in mapping:
                        mapping[local_field] = set()
                    mapping[local_field].add(node)

        return {key: list(value) for key, value in mapping.items()}

    def get_jinja2_trigger_nodes(self) -> dict[ComputedAttributeTarget, list[ComputedAttributeTriggerNode]]:
        working_map: dict[ComputedAttributeTarget, dict[str, ComputedAttributeTriggerNode]] = {}
        for node_kind, registered_computed_attribute in self._computed_jinja2_attribute_map.items():
            for local_field, computed_attribute_targets in registered_computed_attribute.local_fields.items():
                for computed_attribute_target in computed_attribute_targets:
                    if computed_attribute_target not in working_map:
                        working_map[computed_attribute_target] = {}
                    if node_kind not in working_map[computed_attribute_target]:
                        working_map[computed_attribute_target][node_kind] = ComputedAttributeTriggerNode(kind=node_kind)
                    working_map[computed_attribute_target][node_kind].attributes.append(local_field)
                    if computed_attribute_target.kind == node_kind:
                        working_map[computed_attribute_target][node_kind].targets_self = True
            for relationship, computed_attribute_targets in registered_computed_attribute.relationships.items():
                for computed_attribute_target in computed_attribute_targets:
                    if computed_attribute_target not in working_map:
                        working_map[computed_attribute_target] = {}
                    if node_kind not in working_map[computed_attribute_target]:
                        working_map[computed_attribute_target][node_kind] = ComputedAttributeTriggerNode(kind=node_kind)
                    working_map[computed_attribute_target][node_kind].relationships.append(relationship)
                    if computed_attribute_target.kind == node_kind:
                        working_map[computed_attribute_target][node_kind].targets_self = True

        return {
            computed_attribute_target: list(nodes.values()) for computed_attribute_target, nodes in working_map.items()
        }
